import logging
from copy import deepcopy

from pydantic import BaseModel, ValidationError

from labels.model.file import LocationReadCloser
from labels.model.package import Language, Package, PackageType
from labels.model.relationship import Relationship
from labels.model.resolver import Resolver
from labels.parsers.cataloger.generic.parser import Environment
from labels.parsers.cataloger.java.model import JavaArchive, JavaPomProject
from labels.parsers.cataloger.java.package import package_url
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)


class LockFileDependency(BaseModel):
    group: str
    name: str
    version: str
    line: int | None = None


def parse_gradle_lockfile(
    _resolver: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    dependencies: list[LockFileDependency] = []
    packages: list[Package] = []
    for line_number, line in enumerate(reader.read_closer.readlines(), 1):
        if "=" in line and ":" in line:  # To ensure it's a dependency line
            dependency_part = line.split("=")[0]
            group, name, version = dependency_part.split(":")
            dependencies.append(
                LockFileDependency(group=group, name=name, version=version, line=line_number),
            )

    for dependency in dependencies:
        name = dependency.name
        version = dependency.version

        if not name or not version:
            continue

        location = deepcopy(reader.location)
        if location.coordinates:
            location.coordinates.line = dependency.line

        archive = JavaArchive(
            pom_project=JavaPomProject(
                group_id=dependency.group,
                name=name,
                artifact_id=name,
                version=version,
            ),
        )

        try:
            packages.append(
                Package(
                    name=name,
                    version=version,
                    locations=[location],
                    language=Language.JAVA,
                    type=PackageType.JavaPkg,
                    metadata=archive,
                    p_url=package_url(name, version, archive),
                    licenses=[],
                ),
            )
        except ValidationError as ex:
            LOGGER.warning(
                "Malformed package. Required fields are missing or data types are incorrect.",
                extra={
                    "extra": {
                        "exception": format_exception(str(ex)),
                        "location": location.path(),
                    },
                },
            )
            continue

    return packages, []



import logging
import re
from copy import deepcopy

from bs4 import BeautifulSoup
from bs4.element import Tag
from packageurl import PackageURL

from labels.model.file import DependencyType, LocationReadCloser
from labels.model.package import Language, Package, PackageType
from labels.model.relationship import Relationship
from labels.model.resolver import Resolver
from labels.parsers.cataloger.generic.parser import Environment

LOGGER = logging.getLogger(__name__)


def parse_dotnet_config_executable(
    _resolver: Resolver | None,
    _env: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    packages = []
    root = BeautifulSoup(reader.read_closer.read(), features="html.parser")
    net_dep = re.compile(r".NETFramework,Version=v(?P<version>[^\s,]*)")
    net_runtime = root.find("supportedruntime")
    if (
        isinstance(net_runtime, Tag)
        and isinstance(net_runtime.parent, Tag)
        and net_runtime.parent.name == "startup"
        and (runtime_info := net_runtime.get("sku", ""))
        and (version_match := net_dep.match(str(runtime_info)))
    ):
        version = version_match.group("version")

        location = deepcopy(reader.location)
        if location.coordinates:
            location.coordinates.line = net_runtime.sourceline
            location.dependency_type = DependencyType.DIRECT

        packages.append(
            Package(
                name="netframework",
                version=version,
                locations=[location],
                licenses=[],
                type=PackageType.DotnetPkg,
                language=Language.DOTNET,
                metadata=None,
                p_url=PackageURL(
                    type="nuget",
                    namespace="",
                    name="netframework",
                    version=version,
                    qualifiers={},
                    subpath="",
                ).to_string(),
            ),
        )

    return packages, []



import logging
from collections.abc import Generator
from contextlib import suppress
from copy import deepcopy

import requirements
from pydantic import ValidationError
from requirements.requirement import Requirement

from labels.model.file import DependencyType, Location, LocationReadCloser, Scope
from labels.model.package import Language, Package, PackageType
from labels.model.relationship import Relationship
from labels.model.resolver import Resolver
from labels.parsers.cataloger.generic.parser import Environment
from labels.parsers.cataloger.python.model import PythonRequirementsEntry
from labels.parsers.cataloger.python.package import package_url
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)

OPERATOR_ORDER = {"==": 1, "===": 1, "~=": 1, ">=": 2, ">": 2, "<": 3, "<=": 3}


def get_dep_version_range(dep_specs: list[tuple[str, str]]) -> str | None:
    version_range = ""
    ordered_specs = sorted(dep_specs, key=lambda x: OPERATOR_ORDER.get(x[0], 1))
    for operator, version in ordered_specs:
        if operator not in OPERATOR_ORDER:
            return None

        if operator in {"==", "~="}:
            version_range = version
            break
        version_range += f"{operator}{version} "
    return version_range.rstrip()


def get_parsed_dependency(line: str) -> tuple[str, str, Requirement, bool] | None:
    with suppress(Exception):
        parsed_dep = next(iter(requirements.parse(line)))
        if not parsed_dep.specs or not (version := get_dep_version_range(parsed_dep.specs)):
            return None
        is_dev: bool = False
        if parsed_dep.extras and any("dev" in extra.lower() for extra in parsed_dep.extras):
            is_dev = True
        return str(parsed_dep.name), version, parsed_dep, is_dev
    return None


def split_lines_requirements(
    content: str,
) -> Generator[tuple[int, str], None, None]:
    last_line = ""
    line_number = 1
    for index, raw_line in enumerate(content.splitlines(), 1):
        if not last_line:
            line_number = index
        line = trim_requirements_txt_line(raw_line)
        if last_line != "":
            line = last_line + line
            last_line = ""
        if line.endswith("\\"):
            last_line += line.rstrip("\\")
            continue
        if not line:
            continue

        if any(
            (
                line.startswith("-e"),
                line.startswith("-r"),
                line.startswith("--requirements"),
            ),
        ):
            continue

        yield line_number, line


def parse_requirements_txt(
    _resolver: Resolver | None,
    _env: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    try:
        content = reader.read_closer.read()
    except UnicodeDecodeError:
        return [], []

    packages: list[Package] = []
    deps_found = False

    for line_number, line in split_lines_requirements(content):
        parsed_dep = get_parsed_dependency(line)

        if not parsed_dep:
            if not deps_found and line_number > 3:
                return [], []
            continue

        product, version, req, is_dev = parsed_dep
        if not product or not version:
            continue

        deps_found = True
        current_location = prepare_location(reader.location, line_number, is_dev=is_dev)
        p_url = package_url(product, version, None)

        package = create_package(product, version, req, current_location, p_url)
        if package:
            packages.append(package)

    return packages, []


def prepare_location(
    location: Location,
    line_number: int,
    *,
    is_dev: bool = False,
) -> Location:
    current_location = deepcopy(location)
    if current_location.coordinates:
        current_location.dependency_type = DependencyType.DIRECT
        current_location.scope = Scope.DEV if is_dev else Scope.PROD
        current_location.coordinates.line = line_number
    return current_location


def create_package(
    product: str,
    version: str,
    req: Requirement,
    location: Location,
    p_url: str,
) -> Package | None:
    try:
        return Package(
            name=product,
            version=version,
            found_by=None,
            locations=[location],
            language=Language.PYTHON,
            p_url=p_url,
            metadata=PythonRequirementsEntry(
                name=str(req.name),
                extras=sorted(req.extras),
                version_constraint=",".join(f"{s[0]} {s[1]}" for s in req.specs)
                if req.specs
                else "",
                markers=p_url,
            ),
            licenses=[],
            type=PackageType.PythonPkg,
        )
    except ValidationError as ex:
        LOGGER.warning(
            "Malformed package. Required fields are missing or data types are incorrect.",
            extra={
                "extra": {
                    "exception": format_exception(str(ex)),
                    "location": location.path(),
                },
            },
        )
        return None


def remove_trailing_comment(line: str) -> str:
    parts = line.split("#", 1)
    if len(parts) < 2:
        # there aren't any comments
        return line
    return parts[0]


def parse_url(line: str) -> str:
    parts = line.split("@")

    if len(parts) > 1:
        desired_index = -1

        for index, raw_part in enumerate(parts):
            part = "".join([char for char in raw_part if char.isalnum()])

            if part.startswith("git"):
                desired_index = index
                break

        if desired_index != -1:
            return "@".join(parts[desired_index:]).strip()

    return ""


def trim_requirements_txt_line(line: str) -> str:
    line = line.strip()

    return remove_trailing_comment(line)



import logging

from license_expression import ExpressionError, LicenseExpression, get_spdx_licensing
from spdx_tools.spdx.model.actor import Actor, ActorType
from spdx_tools.spdx.model.checksum import Checksum, ChecksumAlgorithm
from spdx_tools.spdx.model.document import Document
from spdx_tools.spdx.model.package import (
    ExternalPackageRef,
    ExternalPackageRefCategory,
    PackagePurpose,
)
from spdx_tools.spdx.model.package import Package as SPDX_Package
from spdx_tools.spdx.model.spdx_no_assertion import SpdxNoAssertion
from spdx_tools.spdx.validation.uri_validators import validate_url

from labels.model.file import Location
from labels.model.package import HealthMetadata, Package
from labels.output.utils import get_author_info, sanitize_name

NOASSERTION = SpdxNoAssertion()
NAMESPACE = "fluid-attacks"

LOGGER = logging.getLogger(__name__)


def is_valid_url(url: str) -> bool:
    return not validate_url(url)


def get_valid_advisory_urls(advisory_urls: list[str]) -> str:
    valid_urls = []
    for url in advisory_urls:
        if url and url != "MANUAL" and is_valid_url(url):
            valid_urls.append(url)
        elif url and url != "MANUAL":
            LOGGER.warning("Invalid advisory URL found and skipped: %s", url)

    return ", ".join(valid_urls) if valid_urls else "https://fluidattacks.com"


def add_vulnerabilities_spdx(package: Package) -> list[ExternalPackageRef]:
    vulnerabilities_refs = []
    if package.advisories:
        for advisory in package.advisories:
            combined_urls = get_valid_advisory_urls(advisory.urls)

            comment_parts = [
                f"Severity: {advisory.severity}",
                f"EPSs: {advisory.epss}",
                f"Score: {advisory.percentile}",
                f"Affected Version: {package.version}",
                f"Affected Version Range: {advisory.version_constraint}"
                if advisory.version_constraint
                else None,
                f"Description: {advisory.description}" if advisory.description else None,
            ]
            comment_text = "; ".join(filter(None, comment_parts))

            vulnerabilities_refs.append(
                ExternalPackageRef(
                    category=ExternalPackageRefCategory.SECURITY,
                    reference_type="advisory",
                    locator=combined_urls,
                    comment=comment_text,
                ),
            )

    return vulnerabilities_refs


def add_authors(
    health_metadata: HealthMetadata | None,
) -> Actor | SpdxNoAssertion:
    if health_metadata is None:
        return NOASSERTION

    authors_info = get_author_info(health_metadata)

    names = []
    emails = []

    for name, email in authors_info:
        if name:
            names.append(name)
        if email:
            emails.append(email)

    concatenated_names = ", ".join(names)
    concatenated_emails = ", ".join(emails) or None

    return Actor(
        actor_type=ActorType.PERSON,
        name=concatenated_names,
        email=concatenated_emails,
    )


def add_locations_external_refs(
    locations: list[Location],
) -> list[ExternalPackageRef]:
    locations_external_refs = []
    for idx, location in enumerate(locations):
        path = location.path()
        coordinates = location.coordinates
        line = coordinates.line if coordinates else None
        layer = coordinates.file_system_id if coordinates else None

        locations_external_refs.append(
            ExternalPackageRef(
                category=ExternalPackageRefCategory.OTHER,
                reference_type=f"{NAMESPACE}:locations:{idx}:path",
                locator=path,
            ),
        )

        if line:
            locations_external_refs.append(
                ExternalPackageRef(
                    category=ExternalPackageRefCategory.OTHER,
                    reference_type=f"{NAMESPACE}:locations:{idx}:line",
                    locator=str(line),
                ),
            )

        if layer:
            locations_external_refs.append(
                ExternalPackageRef(
                    category=ExternalPackageRefCategory.OTHER,
                    reference_type=f"{NAMESPACE}:locations:{idx}:layer",
                    locator=layer,
                ),
            )

    return locations_external_refs


def add_health_metadata_external_refs(
    health_metadata: HealthMetadata,
) -> list[ExternalPackageRef]:
    health_metadata_external_refs = []

    if health_metadata.latest_version:
        health_metadata_external_refs.append(
            ExternalPackageRef(
                category=ExternalPackageRefCategory.OTHER,
                reference_type=(f"{NAMESPACE}:health_metadata:latest_version"),
                locator=health_metadata.latest_version,
            ),
        )

    if health_metadata.latest_version_created_at:
        health_metadata_external_refs.append(
            ExternalPackageRef(
                category=ExternalPackageRefCategory.OTHER,
                reference_type=(f"{NAMESPACE}:health_metadata:latest_version_created_at"),
                locator=str(health_metadata.latest_version_created_at).replace(" ", ""),
            ),
        )

    return health_metadata_external_refs


def add_purl_external_ref(
    package_purl: str,
) -> ExternalPackageRef:
    return ExternalPackageRef(
        category=ExternalPackageRefCategory.PACKAGE_MANAGER,
        reference_type="purl",
        locator=package_purl,
    )


def add_language_external_ref(language: str) -> ExternalPackageRef:
    return ExternalPackageRef(
        category=ExternalPackageRefCategory.OTHER,
        reference_type=f"{NAMESPACE}:language",
        locator=language,
    )


def add_external_refs(package: Package) -> list[ExternalPackageRef]:
    external_refs = []
    health_metadata = package.health_metadata

    external_refs.append(add_language_external_ref(package.language))

    if health_metadata:
        external_refs.extend(add_health_metadata_external_refs(health_metadata))

    external_refs.append(add_purl_external_ref(package.p_url))
    external_refs.extend(add_locations_external_refs(package.locations))
    external_refs.extend(add_vulnerabilities_spdx(package))

    return external_refs


def add_integrity(health_metadata: HealthMetadata | None) -> list[Checksum]:
    if health_metadata:
        integrity = health_metadata.artifact.integrity if health_metadata.artifact else None
        if integrity and integrity.algorithm and integrity.value:
            return [
                Checksum(
                    algorithm=ChecksumAlgorithm[integrity.algorithm.upper()],
                    value=integrity.value,
                ),
            ]
    return []


def add_empty_package(document: Document) -> None:
    document.creation_info.document_comment = (
        "No packages or relationships were found in the resource."
    )

    empty_package = SPDX_Package(
        name="NONE",
        spdx_id="SPDXRef-Package-NONE",
        download_location=NOASSERTION,
        license_concluded=NOASSERTION,
        primary_package_purpose=PackagePurpose.LIBRARY,
    )

    document.packages = [empty_package]
    document.relationships = []


def get_spdx_id(package: Package) -> str:
    name = sanitize_name(package.name)
    pkg_platform = sanitize_name(package.type.value.lower())
    pkg_id = package.id_
    return f"SPDXRef-Package-{pkg_platform}-{name}-{pkg_id}"


def add_license(licenses: list[str]) -> LicenseExpression | SpdxNoAssertion:
    licensing = get_spdx_licensing()
    parsed = NOASSERTION
    try:
        parsed = licensing.parse(" AND ".join(licenses))
    except ExpressionError:
        return NOASSERTION

    return parsed



import logging
from typing import TYPE_CHECKING

from labels.utils.strings import format_exception

if TYPE_CHECKING:  # pragma: no cover
    from collections.abc import ItemsView

from pydantic import ValidationError

from labels.model.file import DependencyType, Location, LocationReadCloser, Scope
from labels.model.indexables import IndexedDict, ParsedValue
from labels.model.package import Language, Package, PackageType
from labels.model.relationship import Relationship
from labels.model.resolver import Resolver
from labels.parsers.cataloger.generic.parser import Environment
from labels.parsers.cataloger.rust.utils import package_url
from labels.parsers.collection.toml import parse_toml_with_tree_sitter

LOGGER = logging.getLogger(__name__)


def _get_location(location: Location, sourceline: int) -> Location:
    if location.coordinates:
        c_upd = {"line": sourceline}
        l_upd = {"coordinates": location.coordinates.model_copy(update=c_upd)}
        location.dependency_type = DependencyType.DIRECT
        return location.model_copy(update=l_upd)
    return location


def _get_version(value: ParsedValue) -> str | None:
    if isinstance(value, str):
        return value
    if not isinstance(value, IndexedDict):
        return None
    if "git" in value:
        repo_url: str = str(value.get("git", ""))
        branch: str = str(value.get("branch", ""))
        if repo_url and branch:
            return f"{repo_url}@{branch}"
    version: str = str(value.get("version", ""))
    return version


def _get_packages(
    reader: LocationReadCloser,
    dependencies: ParsedValue,
    *,
    is_dev: bool,
) -> list[Package]:
    if dependencies is None or not isinstance(dependencies, IndexedDict):
        return []

    packages = []

    items: ItemsView[str, ParsedValue] = dependencies.items()

    for name, value in items:
        version = _get_version(value)
        if not name or not version:
            continue

        location = _get_location(reader.location, dependencies.get_key_position(name).start.line)
        location.scope = Scope.DEV if is_dev else Scope.PROD
        try:
            packages.append(
                Package(
                    name=name,
                    version=version,
                    locations=[location],
                    language=Language.RUST,
                    licenses=[],
                    p_url=package_url(name=name, version=version),
                    type=PackageType.RustPkg,
                ),
            )
        except ValidationError as ex:
            LOGGER.warning(
                "Malformed package. Required fields are missing or data types are incorrect.",
                extra={
                    "extra": {
                        "exception": format_exception(str(ex)),
                        "location": location.path(),
                    },
                },
            )
            continue

    return packages


def parse_cargo_toml(
    _: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    content: IndexedDict[
        str,
        ParsedValue,
    ] = parse_toml_with_tree_sitter(reader.read_closer.read())

    deps: ParsedValue = content.get(
        "dependencies",
    )
    dev_deps: ParsedValue = content.get(
        "dev-dependencies",
    )
    packages = [
        *_get_packages(reader, deps, is_dev=False),
        *_get_packages(reader, dev_deps, is_dev=True),
    ]
    return packages, []



import logging
import os
from itertools import zip_longest
from typing import cast

from packageurl import PackageURL
from pydantic import BaseModel, ValidationError

from labels.model.file import Location, LocationReadCloser
from labels.model.package import Digest, Language, Package, PackageType
from labels.model.relationship import Relationship
from labels.model.release import Release
from labels.model.resolver import Resolver
from labels.parsers.cataloger.generic.parser import Environment
from labels.parsers.cataloger.redhat.rpmdb import open_db
from labels.parsers.cataloger.redhat.rpmdb.package import PackageInfo
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)


class RpmFileRecord(BaseModel):
    path: str
    mode: int
    size: int
    digest: Digest
    username: str
    group_name: str | None
    flags: str


class RpmDBEntry(BaseModel):
    id_: str
    name: str
    version: str
    epoch: int | None
    arch: str
    release: str
    source_rpm: str
    size: int
    vendor: str
    modularitylabel: str
    files: list[RpmFileRecord]


def package_url(  # noqa: PLR0913
    *,
    name: str,
    arch: str | None,
    epoch: int | None,
    source_rpm: str,
    version: str,
    release: str,
    distro: Release | None,
) -> str:
    namespace = ""
    if distro:
        namespace = distro.id_
    qualifiers: dict[str, str] = {}
    if arch:
        qualifiers["arch"] = arch
    if epoch:
        qualifiers["epoch"] = str(epoch)
    if source_rpm:
        qualifiers["upstream"] = source_rpm

    return PackageURL(
        type="rpm",
        namespace=namespace,
        name=name,
        version=f"{version}-{release}",
        qualifiers=qualifiers,
        subpath="",
    ).to_string()


def to_int(value: int | None, default: int = 0) -> int:
    return int(value) if isinstance(value, int) else default


def extract_rmp_file_records(resolver: Resolver, entry: PackageInfo) -> list[RpmFileRecord]:
    records: list[RpmFileRecord] = []
    file_attributes = cast(
        tuple[tuple[str, int, str, int, int, int, str, str], ...],
        zip_longest(
            entry.base_names,
            entry.dir_indexes,
            entry.file_digests,
            entry.file_flags,
            entry.file_modes,
            entry.file_sizes,
            entry.user_names,
            entry.group_names,
        ),
    )

    records.extend(
        record
        for attrs in file_attributes
        if (record := create_rpm_file_record(resolver, entry, attrs))
    )

    return records


def create_rpm_file_record(
    resolver: Resolver,
    entry: PackageInfo,
    attrs: tuple[str, int, str, int, int, int, str, str],
) -> RpmFileRecord | None:
    (
        base_name,
        dir_index,
        file_digest,
        file_flag,
        file_mode,
        file_size,
        file_username,
        file_groupname,
    ) = attrs

    if not base_name or not isinstance(dir_index, int):
        return None

    file_path = os.path.join(str(entry.dir_names[dir_index]), str(base_name))
    file_location = resolver.files_by_path(file_path)

    if not file_location:
        return None

    return RpmFileRecord(
        path=file_path,
        mode=to_int(file_mode, default=0),
        size=to_int(file_size, default=0),
        digest=Digest(
            algorithm="md5" if file_digest else None,
            value=str(file_digest) if file_digest else None,
        ),
        username=str(file_username),
        flags=str(file_flag),
        group_name=str(file_groupname) if file_groupname else None,
    )


def new_redhat_package(
    *,
    entry: PackageInfo,
    resolver: Resolver,
    location: Location,
    env: Environment,
) -> Package | None:
    name = entry.name
    version = entry.version

    if not name or not version:
        return None

    metadata = RpmDBEntry(
        id_="",
        name=name,
        version=version,
        epoch=entry.epoch,
        arch=entry.arch,
        release=entry.release,
        source_rpm=entry.source_rpm,
        vendor=entry.vendor,
        size=entry.size,
        modularitylabel=entry.modularity_label,
        files=extract_rmp_file_records(resolver, entry),
    )
    try:
        return Package(
            name=name,
            version=version,
            locations=[location],
            language=Language.UNKNOWN_LANGUAGE,
            licenses=[entry.license],
            type=PackageType.RpmPkg,
            metadata=metadata,
            p_url=package_url(
                name=name,
                arch=entry.arch,
                epoch=entry.epoch,
                source_rpm=entry.source_rpm,
                version=version,
                release=entry.release,
                distro=env.linux_release,
            ),
        )
    except ValidationError as ex:
        LOGGER.warning(
            "Malformed package. Required fields are missing or data types are incorrect.",
            extra={
                "extra": {
                    "exception": format_exception(str(ex)),
                    "location": location.path(),
                },
            },
        )
        return None


def parse_rpm_db(
    resolver: Resolver,
    env: Environment,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    packages: list[Package] = []

    if not reader.location.coordinates:
        return packages, []

    database = open_db(reader.location.coordinates.real_path)

    if not database:
        return packages, []

    for entry in database.list_packages():
        package = new_redhat_package(
            resolver=resolver,
            entry=entry,
            env=env,
            location=reader.location,
        )
        if package is not None:
            packages.append(package)
    return packages, []



from contextlib import suppress
from datetime import datetime

from bs4 import BeautifulSoup, Tag

from labels.enrichers.alpine.get import get_package_versions_html
from labels.model.package import HealthMetadata, Package
from labels.parsers.cataloger.utils import extract_distro_info


def _get_latest_version_and_latest_version_created_at(
    package: Package,
    distro_version: str | None,
    arch: str | None,
) -> tuple[str, datetime | None] | tuple[None, None]:
    html_content = get_package_versions_html(package.name, distro_version, arch)
    if not html_content:
        return None, None

    parsed_content = BeautifulSoup(html_content, features="html.parser")
    version_items: list[Tag] = list(parsed_content.find_all("td", {"class": "version"}))

    if version_items:
        latest_version = version_items[0].text.strip()
        latest_version_created_at = None
        with suppress(IndexError):
            parent_tr = version_items[0].find_parent("tr")
            if parent_tr and (build_date_tag := parent_tr.find_next("td", {"class": "bdate"})):
                with suppress(ValueError):
                    latest_version_created_at = datetime.fromisoformat(build_date_tag.text.strip())
        return latest_version, latest_version_created_at

    return None, None


def _set_health_metadata(package: Package, arch: str | None, distro_version: str | None) -> None:
    authors = (
        package.metadata if package.metadata and hasattr(package.metadata, "maintainer") else None
    )
    (
        latest_version,
        latest_version_created_at,
    ) = _get_latest_version_and_latest_version_created_at(package, distro_version, arch)

    if not any([latest_version, latest_version_created_at, authors]):
        return

    package.health_metadata = HealthMetadata(
        latest_version=latest_version,
        latest_version_created_at=latest_version_created_at,
        authors=authors.maintainer if authors else None,
    )


def complete_package(
    package: Package,
) -> Package:
    _, distro_version, arch = extract_distro_info(package.p_url)
    _set_health_metadata(package, arch, distro_version)

    return package



from labels.enrichers.dart.get import PubPackage, PubPackageVersion, get_pub_package
from labels.enrichers.utils import infer_algorithm
from labels.model.package import Artifact, Digest, HealthMetadata, Package


def _get_current_package(pub_package: PubPackage, version: str) -> PubPackageVersion | None:
    return next((v for v in pub_package["versions"] if v["version"] == version), None)


def _get_authors(pub_package: PubPackage) -> str | None:
    return next(
        (
            version["pubspec"]["author"]
            for version in reversed(pub_package["versions"])
            if "author" in version["pubspec"]
        ),
        None,
    )


def _get_artifact(current_package: PubPackageVersion) -> Artifact:
    digest_value = current_package["archive_sha256"]
    return Artifact(
        url=current_package["archive_url"],
        integrity=Digest(
            value=digest_value,
            algorithm=infer_algorithm(digest_value),
        ),
    )


def _set_health_metadata(
    package: Package,
    pub_package: PubPackage,
    current_package: PubPackageVersion | None,
) -> None:
    package.health_metadata = HealthMetadata(
        latest_version=pub_package["latest"]["version"],
        latest_version_created_at=pub_package["latest"]["published"],
        authors=_get_authors(pub_package),
        artifact=_get_artifact(current_package) if current_package else None,
    )


def complete_package(package: Package) -> Package:
    pub_package = get_pub_package(package.name)
    if not pub_package:
        return package

    current_package = _get_current_package(pub_package, package.version)

    _set_health_metadata(package, pub_package, current_package)

    return package



from labels.enrichers.php.get import PackagistPackageInfo, get_composer_package
from labels.enrichers.utils import infer_algorithm
from labels.model.package import Artifact, Digest, HealthMetadata, Package
from labels.utils.licenses.validation import validate_licenses


def _get_author(composer_package: PackagistPackageInfo) -> str | None:
    if not composer_package.get("authors"):
        return None

    authors: list[str] = []
    authors_dict = composer_package["authors"]
    for author_item in authors_dict:
        author: str = author_item["name"]
        if "email" in author_item:
            author_email = author_item["email"]
            author += f" <{author_email}>"
        authors.append(author)

    return ", ".join(authors)


def _set_health_metadata(
    package: Package,
    composer_package: PackagistPackageInfo,
    current_package: PackagistPackageInfo | None,
) -> None:
    package.health_metadata = HealthMetadata(
        latest_version=composer_package["version"],
        latest_version_created_at=composer_package["time"],
        artifact=_get_artifact_metadata(current_package),
        authors=_get_author(composer_package),
    )


def _get_artifact_metadata(
    current_package: PackagistPackageInfo | None,
) -> Artifact | None:
    if current_package:
        dist_info = current_package.get("dist")
        if isinstance(dist_info, dict) and isinstance(dist_info.get("url"), str):
            digest_value = dist_info.get("shasum") or None
            return Artifact(
                url=dist_info["url"],
                integrity=Digest(
                    algorithm=infer_algorithm(digest_value),
                    value=digest_value,
                ),
            )
    return None


def complete_package(package: Package) -> Package:
    current_package = get_composer_package(package.name, package.version)
    # The p2/$vendor/$package.json file contains only tagged releases, not development versions
    composer_package = get_composer_package(package.name)

    if not composer_package:
        return package

    _set_health_metadata(package, composer_package, current_package)

    package.licenses = validate_licenses(composer_package["license"])

    return package



from __future__ import (
    annotations,
)

import struct


class HashOffPageEntry:
    def __init__(self, page_type: int, unused: int, page_no: int, length: int) -> None:
        self.page_type = page_type  # uint8
        self.unused = unused  # 3 bytes
        self.page_no = page_no  # uint32
        self.length = length  # uint32

    @classmethod
    def from_bytes(cls, data: bytes, *, swapped: bool) -> HashOffPageEntry:
        expected_size = 12  # Total size in bytes
        if len(data) < expected_size:
            error_msg = (
                f"Data too short, expected at least {expected_size} bytes, got {len(data)} bytes"
            )
            raise ValueError(error_msg)

        # Determine the byte order
        byte_order_str = ">" if swapped else "<"  # '>' for big-endian, '<' for little-endian

        # Struct format string
        fmt = f"{byte_order_str}B3sII"  # Format: PageType(uint8),
        # Unused(3 bytes), PageNo(uint32), Length(uint32)

        try:
            unpacked_data = struct.unpack(fmt, data[:expected_size])
        except struct.error as exc:
            error_msg = f"Failed to unpack HashOffPageEntry: {exc}"
            raise ValueError(error_msg) from exc

        page_type, unused, page_no, length = unpacked_data

        return cls(page_type, unused, page_no, length)

    def __repr__(self) -> str:
        return (
            f"HashOffPageEntry("
            f"PageType={self.page_type}, Unused={self.unused}, "
            f"PageNo={self.page_no}, Length={self.length})"
        )


def parse_hash_off_page_entry(data: bytes, *, swapped: bool) -> HashOffPageEntry:
    return HashOffPageEntry.from_bytes(data, swapped=swapped)



from __future__ import (
    annotations,
)

import struct

from pydantic import (
    BaseModel,
)

from labels.parsers.cataloger.redhat.rpmdb.berkeley.constants import (
    HASH_MAGIC_NUMBER,
    HASH_MAGIC_NUMBER_BE,
    HASH_METADATA_PACKAGE_TYPE,
)
from labels.parsers.cataloger.redhat.rpmdb.berkeley.generic_page import (
    GenericMetadataPage,
)
from labels.utils.exceptions import (
    InvalidMetadataError,
)


class HashMetadata(BaseModel):
    """Represents the Hash Metadata in the RPM database."""

    generic_metadata_page: GenericMetadataPage
    max_bucket: int
    high_mask: int
    low_mask: int
    fill_factor: int
    num_keys: int
    char_key_hash: int

    @classmethod
    def from_bytes(cls, data: bytes, byte_order: str) -> HashMetadata:
        expected_size = 96  # Up to CharKeyHash (bytes 0-95)
        if len(data) < expected_size:
            error_msg = (
                f"Data too short, expected at least {expected_size} bytes, got {len(data)} bytes"
            )
            raise ValueError(error_msg)

        # Parse the GenericMetadataPage part
        generic_metadata = GenericMetadataPage.from_bytes(data[:72], byte_order)

        # Parse the additional fields (MaxBucket to CharKeyHash)
        fmt = f"{'>' if byte_order == 'big' else '<'}6I"
        additional_fields_size = struct.calcsize(fmt)
        try:
            unpacked_data = struct.unpack(fmt, data[72 : 72 + additional_fields_size])
        except struct.error as exc:
            error_msg = f"Failed to unpack HashMetadata additional fields: {exc}"
            raise ValueError(error_msg) from exc

        (
            max_bucket,
            high_mask,
            low_mask,
            fill_factor,
            num_keys,
            char_key_hash,
        ) = unpacked_data

        return cls(
            generic_metadata_page=generic_metadata,
            max_bucket=max_bucket,
            high_mask=high_mask,
            low_mask=low_mask,
            fill_factor=fill_factor,
            num_keys=num_keys,
            char_key_hash=char_key_hash,
        )

    @classmethod
    def validate(cls, value: HashMetadata) -> None:  # type: ignore
        """Validate the HashMetadata.

        :raises InvalidMetadata: If validation fails.
        """
        if value.generic_metadata_page.magic != HASH_MAGIC_NUMBER:
            error_msg = f"Unexpected DB magic number: {hex(value.generic_metadata_page.magic)}"
            raise InvalidMetadataError(error_msg)

        if value.generic_metadata_page.page_type != HASH_METADATA_PACKAGE_TYPE:
            raise InvalidMetadataError(
                "Unexpected page type: " + str(value.generic_metadata_page.page_type),
            )


class HashMetadataPage:
    """Represents a Hash Metadata Page with endian information."""

    def __init__(self, hash_metadata: HashMetadata, *, swapped: bool) -> None:
        self.hash_metadata = hash_metadata
        self.swapped = swapped  # Indicates if the byte order is swapped (big-endian)

    @classmethod
    def from_bytes(cls, data: bytes) -> HashMetadataPage:
        """Parse a HashMetadataPage from a bytes object.

        :param data: The bytes object containing the page data.
        :return: An instance of HashMetadataPage.
        :raises ValueError: If the data is invalid or parsing fails.
        """
        swapped = False
        byte_order = "little"

        # First attempt to parse using little-endian byte order
        hash_metadata = HashMetadata.from_bytes(data, byte_order)

        if hash_metadata.generic_metadata_page.magic == HASH_MAGIC_NUMBER_BE:
            # Re-read the GenericMetadataPage using big-endian byte order
            swapped = True
            byte_order = "big"

            # Re-parse only the GenericMetadataPage portion
            generic_metadata = GenericMetadataPage.from_bytes(data[:72], byte_order)

            # Create a new HashMetadata instance with the corrected
            # GenericMetadataPage
            metadata = HashMetadata(
                generic_metadata_page=generic_metadata,
                max_bucket=hash_metadata.max_bucket,
                high_mask=hash_metadata.high_mask,
                low_mask=hash_metadata.low_mask,
                fill_factor=hash_metadata.fill_factor,
                num_keys=hash_metadata.num_keys,
                char_key_hash=hash_metadata.char_key_hash,
            )
            hash_metadata = metadata

        # Validate the metadata
        HashMetadata.validate(hash_metadata)

        return cls(hash_metadata, swapped=swapped)

    def __repr__(self) -> str:
        return f"HashMetadataPage(hash_metadata={self.hash_metadata}, swapped={self.swapped})"


def parse_hash_metadata_page(data: bytes) -> HashMetadataPage:
    """Parse a HashMetadataPage from a bytes object.

    :param data: A bytes object containing the metadata page data.
    :type data: bytes
    :return: An instance of HashMetadataPage.
    :rtype: HashMetadataPage
    :raises ValueError | InvalidMetadataError: If the data is invalid or parsing
        fails.
    """
    return HashMetadataPage.from_bytes(data)



import logging

from packageurl import PackageURL
from pydantic import BaseModel, ValidationError

from labels.model.file import Location
from labels.model.package import Digest, Language, Package, PackageType
from labels.model.release import Release
from labels.utils.licenses.validation import validate_licenses
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)


class ApkFileRecord(BaseModel):
    path: str
    owner_uid: str | None = None
    owner_gid: str | None = None
    permissions: str | None = None
    digest: Digest | None = None


class ApkDBEntry(BaseModel):
    package: str
    origin_package: str | None
    maintainer: str | None
    version: str
    architecture: str | None
    url: str | None
    description: str | None
    size: str
    installed_size: str | None
    dependencies: list[str]
    provides: list[str]
    checksum: str | None
    git_commit: str | None
    files: list[ApkFileRecord]


class ParsedData(BaseModel):
    apk_db_entry: ApkDBEntry
    license: str | None


def package_url(entry: ApkDBEntry, distro: Release | None) -> str:
    qualifiers = {"arch": entry.architecture or ""} if entry else {}

    if entry and entry.origin_package != entry.package and entry.origin_package:
        qualifiers["upstream"] = entry.origin_package
    distro_qualifiers = []

    if distro and distro.id_:
        qualifiers["distro_id"] = distro.id_
        distro_qualifiers.append(distro.id_)

    if distro and distro.version_id:
        qualifiers["distro_version_id"] = distro.version_id
        distro_qualifiers.append(distro.version_id)
    elif distro and distro.build_id:
        distro_qualifiers.append(distro.build_id)

    if distro_qualifiers:
        qualifiers["distro"] = "-".join(distro_qualifiers)

    return PackageURL(
        type="apk",
        namespace=distro.id_.lower() if distro and distro.id_ else "",
        name=entry.package,
        version=entry.version,
        qualifiers=qualifiers,
        subpath="",
    ).to_string()


def new_package(
    data: ParsedData,
    release: Release | None,
    db_location: Location,
) -> Package | None:
    name = data.apk_db_entry.package
    version = data.apk_db_entry.version

    if not name or not version:
        return None

    try:
        return Package(
            name=name,
            version=version,
            locations=[db_location],
            licenses=validate_licenses(data.license.split(" ")) if data.license else [],
            p_url=package_url(data.apk_db_entry, release),
            type=PackageType.ApkPkg,
            metadata=data.apk_db_entry,
            found_by=None,
            health_metadata=None,
            language=Language.UNKNOWN_LANGUAGE,
        )
    except ValidationError as ex:
        LOGGER.warning(
            "Malformed package. Required fields are missing or data types are incorrect.",
            extra={
                "extra": {
                    "exception": format_exception(str(ex)),
                    "location": db_location.path(),
                },
            },
        )
        return None



from typing import TypedDict

from pydantic import BaseModel


class OsReleaseDict(TypedDict, total=False):
    PRETTY_NAME: str
    NAME: str
    ID: str
    ID_LIKE: str
    VERSION: str
    VERSION_ID: str
    VERSION_CODENAME: str
    BUILD_ID: str
    IMAGE_ID: str
    IMAGE_VERSION: str
    VARIANT: str
    VARIANT_ID: str
    HOME_URL: str
    SUPPORT_URL: str
    BUG_REPORT_URL: str
    PRIVACY_POLICY_URL: str
    CPE_NAME: str
    SUPPORT_END: str


class Release(BaseModel):
    id_: str
    version_id: str
    name: str | None = None
    pretty_name: str | None = None
    version: str | None = None
    id_like: list[str] | None = None
    version_code_name: str | None = None
    build_id: str | None = None
    image_id: str | None = None
    image_version: str | None = None
    variant: str | None = None
    variant_id: str | None = None
    home_url: str | None = None
    support_url: str | None = None
    bug_report_url: str | None = None
    privacy_policy_url: str | None = None
    cpe_name: str | None = None
    support_end: str | None = None

    def __str__(self) -> str:
        if self.pretty_name:
            return self.pretty_name
        if self.name:
            return self.name
        if self.version:
            return f"{self.id_} {self.version}"
        if self.version_id != "":
            return f"{self.id_} {self.version_id}"
        return f"{self.id_} {self.build_id or ''}"



import logging
from copy import deepcopy
from typing import cast

from pydantic import ValidationError

from labels.model.file import DependencyType, LocationReadCloser, Scope
from labels.model.indexables import IndexedDict, ParsedValue
from labels.model.package import Language, Package, PackageType
from labels.model.relationship import Relationship
from labels.model.resolver import Resolver
from labels.parsers.cataloger.generic.parser import Environment
from labels.parsers.cataloger.javascript.model import NpmPackage
from labels.parsers.cataloger.javascript.package import package_url
from labels.parsers.collection.json import parse_json_with_tree_sitter
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)


def _create_package(
    package_json: IndexedDict[str, ParsedValue],
    reader: LocationReadCloser,
    package_name: str,
    specifier: str,
    *,
    is_dev: bool,
) -> Package | None:
    current_location = deepcopy(reader.location)
    current_location.scope = Scope.DEV if is_dev else Scope.PROD
    dependencies_key = "devDependencies" if is_dev else "dependencies"
    pkg: IndexedDict[str, ParsedValue] = cast(
        IndexedDict[str, ParsedValue],
        package_json[dependencies_key],
    )
    if current_location.coordinates:
        current_location.coordinates.line = pkg.get_key_position(package_name).start.line
        current_location.dependency_type = DependencyType.DIRECT
    try:
        return Package(
            name=package_name,
            version=specifier,
            type=PackageType.NpmPkg,
            language=Language.JAVASCRIPT,
            licenses=[],
            locations=[current_location],
            p_url=package_url(package_name, specifier),
            metadata=NpmPackage(
                name=package_name,
                version=specifier,
                is_dev=is_dev,
            ),
            is_dev=is_dev,
        )
    except ValidationError as ex:
        LOGGER.warning(
            "Malformed package. Required fields are missing or data types are incorrect.",
            extra={
                "extra": {
                    "exception": format_exception(str(ex)),
                    "location": current_location.path(),
                },
            },
        )
        return None


def parse_package_json(
    _: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    package_json: IndexedDict[str, ParsedValue] = cast(
        IndexedDict[str, ParsedValue],
        parse_json_with_tree_sitter(reader.read_closer.read()),
    )
    deps: ParsedValue = package_json.get("dependencies", IndexedDict())
    if not isinstance(deps, IndexedDict):
        LOGGER.warning("No deps found in package JSON")
        return ([], [])

    packages = []
    for package_name, specifier in deps.items():
        if not package_name or not specifier:
            continue

        package = _create_package(
            package_json,
            reader,
            package_name,
            str(specifier),
            is_dev=False,
        )
        if package:
            packages.append(package)

    dev_deps: ParsedValue = package_json.get("devDependencies", IndexedDict())
    if not isinstance(dev_deps, IndexedDict):
        LOGGER.warning("No dev deps found in package JSON")
        return ([], [])
    for package_name, specifier in dev_deps.items():
        if not package_name or not specifier:
            continue

        package = _create_package(
            package_json,
            reader,
            package_name,
            str(specifier),
            is_dev=True,
        )
        if package:
            packages.append(package)

    return packages, []



import logging
from copy import deepcopy
from typing import cast

from labels.model.file import LocationReadCloser
from labels.model.indexables import IndexedDict, IndexedList, ParsedValue
from labels.model.package import Package
from labels.model.relationship import Relationship
from labels.model.resolver import Resolver
from labels.parsers.cataloger.generic.parser import Environment
from labels.parsers.cataloger.swift.package import new_swift_package_manager_package
from labels.parsers.collection.json import parse_json_with_tree_sitter

LOGGER = logging.getLogger(__name__)
EMPTY_DICT: IndexedDict[str, ParsedValue] = IndexedDict()
EMPTY_LIST: IndexedList[ParsedValue] = IndexedList()


def _get_name_and_version(
    pin: ParsedValue,
) -> tuple[str, str, IndexedDict[str, ParsedValue]] | None:
    if not isinstance(pin, IndexedDict):
        return None
    state: ParsedValue = pin.get("state", EMPTY_DICT)
    name = pin.get("identity")
    if not isinstance(state, IndexedDict):
        return None
    version = state.get("version")

    if not name or not version or not isinstance(version, str) or not isinstance(name, str):
        return None

    return name, version, state


def parse_package_resolved(
    _: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    package_resolved: IndexedDict[str, ParsedValue] = cast(
        "IndexedDict[str, ParsedValue]",
        parse_json_with_tree_sitter(reader.read_closer.read()),
    )

    packages: list[Package] = []
    relationships: list[Relationship] = []
    package_resolved_pins: ParsedValue = package_resolved.get("pins", EMPTY_LIST)

    if isinstance(package_resolved_pins, IndexedList):
        for pin in package_resolved_pins:
            if not isinstance(pin, IndexedDict):
                continue
            info = _get_name_and_version(pin)
            if not info:
                continue
            new_location = deepcopy(reader.location)
            if new_location.coordinates:
                new_location.coordinates.line = pin.get_key_position("identity").start.line

            if pkg := new_swift_package_manager_package(
                name=info[0],
                version=info[1],
                source_url=pin.get("location"),
                revision=info[2].get("revision"),
                location=new_location,
            ):
                packages.append(pkg)

    return packages, relationships



import re
from copy import deepcopy
from typing import cast

from labels.model.file import DependencyType, Location, LocationReadCloser, Scope
from labels.model.package import Language, Package, PackageType
from labels.model.relationship import Relationship
from labels.model.resolver import Resolver
from labels.parsers.cataloger.elixir.utils import package_url
from labels.parsers.cataloger.generic.parser import Environment

MIX_DEP: re.Pattern[str] = re.compile(
    r"\{:(?P<dep>[\w]*),\s\"~>\s(?P<version>[\d.]+)\".+",
)


def _get_location(
    reader: LocationReadCloser,
    line_number: int,
    *,
    is_dev: bool,
) -> Location:
    location = deepcopy(reader.location)
    location.scope = Scope.DEV if is_dev else Scope.PROD
    if location.coordinates:
        location.coordinates.line = line_number
        location.dependency_type = DependencyType.DIRECT
    return location


def parse_mix_exs(
    _: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    packages: list[Package] = []
    relationships: list[Relationship] = []
    is_line_deps = False

    for line_number, raw_line in enumerate(
        reader.read_closer.read().splitlines(),
        1,
    ):
        line = raw_line.strip()
        if line == "defp deps do":
            is_line_deps = True
        elif is_line_deps:
            if line == "end":
                break
            if matched := MIX_DEP.match(line):
                is_dev = ":dev" in line
                pkg_name = cast(str, matched.group("dep"))
                pkg_version = cast(str, matched.group("version"))
                location = _get_location(reader, line_number, is_dev=is_dev)

                packages.append(
                    Package(
                        name=pkg_name,
                        version=pkg_version,
                        type=PackageType.HexPkg,
                        locations=[location],
                        p_url=package_url(pkg_name, pkg_version),
                        metadata=None,
                        language=Language.ELIXIR,
                        licenses=[],
                        is_dev=is_dev,
                    ),
                )
    return packages, relationships



from typing import TextIO

from labels.model.package import Digest
from labels.parsers.cataloger.debian.model import DpkgFileRecord


def parse_dpkg_md5_info(reader: TextIO) -> list[DpkgFileRecord]:
    result: list[DpkgFileRecord] = []
    for raw_line in reader:
        line = raw_line.rstrip("\n")

        fields = line.split("  ", 2)
        if len(fields) != 2:
            continue
        path = fields[1].strip()
        if not path.startswith("/"):
            path = f"/{path}"
        result.append(
            DpkgFileRecord(
                path=path,
                digest=Digest(algorithm="md5", value=fields[0].strip()),
            ),
        )
    return result


def _split_lines(reader: TextIO | str) -> list[str]:
    # Convert the input reader to a list of lines
    return reader.splitlines() if isinstance(reader, str) else reader.readlines()


def _process_line(line: str) -> tuple[str, Digest | None]:
    # Process a single line to extract the file path and digest
    line = line.rstrip("\n").strip()
    if not line:
        return "", None

    fields = line.split(" ", 2)
    path = fields[0].strip() if len(fields) >= 1 else ""
    if path and not path.startswith("/"):
        path = f"/{path}"

    digest = Digest(algorithm="md5", value=fields[1].strip()) if len(fields) >= 2 else None
    return path, digest


def _create_record(path: str, digest: Digest | None) -> DpkgFileRecord | None:
    # Create a DpkgFileRecord object if the path is valid
    if not path:
        return None
    record = DpkgFileRecord(path=path, is_config_file=True)
    if digest:
        record.digest = digest
    return record


def parse_dpkg_conffile_info(
    reader: TextIO | str,
) -> list[DpkgFileRecord]:
    # Parse the dpkg conffile information from the reader
    result: list[DpkgFileRecord] = []
    for line in _split_lines(reader):
        path, digest = _process_line(line)
        record = _create_record(path, digest)
        if record:
            result.append(record)
    return result



import logging
from collections.abc import (
    Iterator,
)
from copy import deepcopy
from typing import cast

from more_itertools import flatten
from packageurl import PackageURL
from pydantic import ValidationError

from labels.model.file import LocationReadCloser
from labels.model.indexables import IndexedDict, ParsedValue
from labels.model.package import Language, Package, PackageType
from labels.model.relationship import Relationship, RelationshipType
from labels.model.resolver import Resolver
from labels.parsers.cataloger.generic.parser import Environment
from labels.parsers.collection.json import parse_json_with_tree_sitter
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)

EMPTY_DICT: IndexedDict[str, ParsedValue] = IndexedDict()


def _get_relationships(
    packages: list[Package],
    dependencies: dict[str, ParsedValue],
) -> list[Relationship]:
    result: list[Relationship] = []
    for package_name, depens in dependencies.items():
        if not isinstance(depens, IndexedDict):
            continue
        deps_gen = (
            p
            for p in packages
            for dep_name in depens
            if dep_name == p.name and isinstance(depens, IndexedDict)
        )
        current_package = next((p for p in packages if p.name == package_name), None)
        if current_package is not None:
            result.extend(
                Relationship(
                    from_=p,
                    to_=current_package,
                    type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                )
                for p in deps_gen
            )
    return result


def parse_dotnet_deps_json(
    _resolver: Resolver | None,
    _env: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    package_json = cast(
        IndexedDict[str, ParsedValue],
        parse_json_with_tree_sitter(reader.read_closer.read()),
    )
    packages: list[Package] = []
    targets: ParsedValue = package_json.get("targets", IndexedDict())
    dependencies: dict[str, ParsedValue] = {}
    if not isinstance(targets, IndexedDict):
        LOGGER.warning("No targets found in package JSON")
        return ([], [])
    for package_key, package_value in cast(
        Iterator[IndexedDict[str, ParsedValue]],
        flatten(x.items() for x in targets.values() if isinstance(x, IndexedDict)),
    ):
        if not isinstance(package_key, str) or "/" not in package_key:
            continue
        package_name, version = package_key.split("/", 1)
        if not isinstance(package_value, IndexedDict):
            continue
        location = deepcopy(reader.location)
        pos = package_value.position
        if hasattr(location, "coordinates") and location.coordinates:
            location.coordinates.line = pos.start.line
        dependencies[package_name] = package_value.get("dependencies", EMPTY_DICT)
        try:
            pkg = Package(
                name=package_name,
                version=version,
                locations=[location],
                licenses=[],
                type=PackageType.DotnetPkg,
                language=Language.DOTNET,
                metadata=None,
                p_url=PackageURL(
                    type="nuget",
                    namespace="",
                    name=package_name,
                    version=version,
                    qualifiers={},
                    subpath="",
                ).to_string(),
            )
            packages.append(pkg)
        except ValidationError as ex:
            LOGGER.warning(
                "Malformed package. Required fields are missing or data types are incorrect.",
                extra={
                    "extra": {
                        "exception": format_exception(str(ex)),
                        "location": location.path(),
                    },
                },
            )
            continue

    relationships: list[Relationship] = _get_relationships(packages, dependencies)
    return packages, relationships



import logging
import multiprocessing
import traceback
from collections.abc import Callable
from typing import cast

import reactivex
from reactivex import Observable
from reactivex import operators as ops
from reactivex.abc import ObserverBase, SchedulerBase
from reactivex.scheduler import ThreadPoolScheduler

from labels.model.package import Package
from labels.model.relationship import Relationship, RelationshipType
from labels.model.resolver import Resolver
from labels.parsers.cataloger.generic.cataloger import Request, Task, execute_parsers
from labels.parsers.cataloger.generic.parser import Environment
from labels.parsers.cataloger.handle import handle_parser
from labels.parsers.cataloger.python.model import PythonPackage
from labels.parsers.operations.utils import identify_release

LOGGER = logging.getLogger(__name__)


def strip_version_specifier(item: str) -> str:
    # Define the characters that indicate the start of a version specifier
    specifiers = "[(<>="

    # Find the index of the first occurrence of any specifier character
    index = next((i for i, char in enumerate(item) if char in specifiers), None)

    # If no specifier character is found, return the original string
    if index is None:
        return item.strip()

    # Return the substring up to the first specifier character, stripped of
    # leading/trailing whitespace
    return item[:index].strip()


def handle_relationships(packages: list[Package]) -> list[Relationship]:
    relationships: list[Relationship] = []
    for package in packages:
        match package.found_by:
            case "python-installed-package-cataloger":
                python_package: PythonPackage = cast(PythonPackage, package.metadata)
                for dep in python_package.dependencies if python_package.dependencies else []:
                    dep_name = strip_version_specifier(dep)
                    if dep_package := next((x for x in packages if x.name == dep_name), None):
                        relationships.append(
                            Relationship(
                                from_=dep_package,
                                to_=package,
                                type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                            ),
                        )
    return relationships


def gen_location_tasks(
    resolver: Resolver,
) -> Callable[[Observable[Request]], Observable]:
    def _handle(source: Observable[Request]) -> Observable:
        def subscribe(
            observer: ObserverBase[Task],
            scheduler: SchedulerBase | None = None,
        ) -> reactivex.abc.DisposableBase:
            def on_next(value: Request) -> None:
                try:
                    locations = resolver.files_by_path(value.real_path)
                    for location in locations:
                        observer.on_next(
                            Task(
                                location=location,
                                parser=value.parser,
                                parser_name=value.parser_name,
                            ),
                        )
                except (
                    Exception  # noqa: BLE001
                ) as ex:
                    observer.on_error(ex)

            return source.subscribe(
                on_next,
                observer.on_error,
                observer.on_completed,
                scheduler=scheduler,
            )

        return reactivex.create(subscribe)

    return _handle


def log_and_continue(e: Exception, file_item: str) -> Observable[None]:
    LOGGER.error(
        "Error found while resolving packages of %s: %s: %s",
        file_item,
        str(e),
        traceback.format_exc(),
    )
    return reactivex.empty()


def process_file_item(
    file_item: str,
    resolver: Resolver,
    pool_scheduler: ThreadPoolScheduler,
) -> Observable[tuple[list[Package], list[Relationship]]]:
    def factory(_: SchedulerBase | None = None) -> Observable[str]:
        return reactivex.just(file_item)

    return reactivex.defer(factory).pipe(
        handle_parser(scheduler=pool_scheduler),
        gen_location_tasks(resolver),
        execute_parsers(resolver, Environment(linux_release=identify_release(resolver))),
        ops.catch(lambda e, _: log_and_continue(e, file_item)),
    )


def package_operations_factory(
    resolver: Resolver,
) -> tuple[list[Package], list[Relationship]]:
    observer: Observable[str] = reactivex.from_iterable(resolver.walk_file())
    result_packages: list[Package] = []
    result_relations: list[Relationship] = []
    completed_event = multiprocessing.Event()
    errors = []

    def on_completed() -> None:
        completed_event.set()

    def on_error(error: Exception) -> None:
        errors.append(error)
        on_completed()

    def on_next(value: tuple[list[Package], list[Relationship]]) -> None:
        packages, relations = value
        result_packages.extend(packages)
        result_relations.extend(relations)

    optimal_thread_count = multiprocessing.cpu_count()
    pool_scheduler = ThreadPoolScheduler(optimal_thread_count)

    final_obs: Observable[tuple[list[Package], list[Relationship]]] = observer.pipe(
        ops.map(
            lambda file_item: process_file_item(file_item, resolver, pool_scheduler),  # type: ignore[arg-type]
        ),
        ops.merge(max_concurrent=optimal_thread_count),
    )
    final_obs.subscribe(on_next=on_next, on_error=on_error, on_completed=on_completed)

    completed_event.wait()
    result_relations.extend(handle_relationships(result_packages))

    return result_packages, result_relations



import os

import tree_sitter

GRAMMARS: dict[str, str] = {
    # Grammars required because not available on pypi as of 12/24
    "swift": os.environ["grammarSwift"],
}


def main() -> None:
    out: str = os.environ["out"]
    os.makedirs(out)

    for grammar, src in GRAMMARS.items():
        path = os.path.join(out, f"{grammar}.so")
        tree_sitter.Language.build_library(path, [src])


if __name__ == "__main__":
    main()



from copy import deepcopy
from typing import cast

from packageurl import PackageURL

from labels.model.file import LocationReadCloser
from labels.model.indexables import IndexedDict, IndexedList, ParsedValue
from labels.model.package import Language, Package, PackageType
from labels.model.relationship import Relationship
from labels.model.resolver import Resolver
from labels.parsers.cataloger.generic.parser import Environment
from labels.parsers.collection.yaml import parse_yaml_with_tree_sitter


def package_url(name: str, version: str) -> str:
    return PackageURL(  # type: ignore
        type="github",
        namespace="",
        name=name,
        version=version,
        qualifiers=None,
        subpath="",
    ).to_string()


def extract_dep(step: IndexedDict[str, ParsedValue]) -> tuple[str, int] | None:
    uses = step.get("uses")
    if not isinstance(uses, str):
        return None
    if "name" in step:
        line_number = step.get_key_position("name").start.line
    else:
        line_number = step.get_key_position("uses").start.line
    return uses, line_number


def _get_deps(jobs: IndexedDict[str, ParsedValue]) -> list[tuple[str, int]]:
    deps: list[tuple[str, int]] = []
    for job in jobs.values():
        if isinstance(job, IndexedDict):
            steps = job.get("steps")
            if isinstance(steps, IndexedList):
                deps.extend(
                    filter(
                        None,
                        (extract_dep(step) for step in steps if isinstance(step, IndexedDict)),
                    ),
                )
    return deps


def parse_github_actions_deps(
    _: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    packages: list[Package] = []
    parsed_content = cast(
        "IndexedDict[str, ParsedValue]",
        parse_yaml_with_tree_sitter(reader.read_closer.read()),
    )
    if not parsed_content:
        return packages, []
    jobs = parsed_content.get("jobs")
    if not jobs or not isinstance(jobs, IndexedDict):
        return packages, []

    for dep, line_number in _get_deps(jobs):
        dep_info = dep.rsplit("@", 1)
        location = deepcopy(reader.location)
        if location.coordinates:
            location.coordinates.line = line_number
        if len(dep_info) == 2:
            packages.append(
                Package(
                    name=dep_info[0],
                    version=dep_info[1],
                    language=Language.GITHUB_ACTIONS,
                    licenses=[],
                    locations=[location],
                    type=PackageType.GithubActionPkg,
                    metadata=None,
                    p_url=package_url(dep_info[0], dep_info[1]),
                ),
            )
    return packages, []



import hashlib
import json
from datetime import datetime
from enum import Enum, StrEnum

from pydantic import BaseModel, Field, field_validator

from labels.model.advisories import Advisory
from labels.model.file import Location


class Platform(StrEnum):
    ALPM = "ALPM"
    APK = "ALPINE"
    BINARY = "BINARY"
    CABAL = "CABAL"
    CARGO = "CARGO"
    COMPOSER = "COMPOSER"
    CONAN = "CONAN"
    CRAN = "CRAN"
    DEBIAN = "DEBIAN"
    ERLANG = "ERLANG"
    GEM = "GEM"
    GITHUB_ACTIONS = "GITHUB_ACTIONS"
    GO = "GO"
    GRAALVM_NATIVE_IMAGE = "GRAALVM_NATIVE_IMAGE"
    JENKINS_PLUGIN = "JENKINS_PLUGIN"
    LINUX_KERNEL = "LINUX_KERNEL"
    LINUX_KERNEL_MODULE = "LINUX_KERNEL_MODULE"
    MAVEN = "MAVEN"
    MSRC_KB = "MSRC_KB"
    NIXPKGS = "NIXPKGS"
    NPM = "NPM"
    NUGET = "NUGET"
    PIP = "PIP"
    PORTAGE = "PORTAGE"
    PUB = "PUB"
    RPM = "RPM"
    SWIFT = "SWIFT"
    UNKNOWN = "UNKNOWN"
    WORDPRESS_PLUGIN = "WORDPRESS_PLUGIN"


class Language(StrEnum):
    UNKNOWN_LANGUAGE = "unknown_language"
    CPP = "c++"
    DART = "dart"
    DOTNET = "dotnet"
    ELIXIR = "elixir"
    ERLANG = "erlang"
    GO = "go"
    GITHUB_ACTIONS = "github_actions"
    HASKELL = "haskell"
    JAVA = "java"
    JAVASCRIPT = "javascript"
    PHP = "php"
    PYTHON = "python"
    R = "R"
    RUBY = "ruby"
    RUST = "rust"
    SWIFT = "swift"

    def get_platform_value(self) -> str | None:
        """Deprecate this method; use PackageType.get_platform_value() instead.

        This method will be removed in the future. Please update your code to use
        PackageType.get_platform_value().
        """
        # DEPRECATED: Use PackageType.get_platform_value() instead.
        language_to_platform = {
            Language.CPP: Platform.CONAN.value,
            Language.DART: Platform.PUB.value,
            Language.DOTNET: Platform.NUGET.value,
            Language.ELIXIR: Platform.ERLANG.value,
            Language.ERLANG: Platform.ERLANG.value,
            Language.GO: Platform.GO.value,
            Language.HASKELL: Platform.CABAL.value,
            Language.JAVA: Platform.MAVEN.value,
            Language.JAVASCRIPT: Platform.NPM.value,
            Language.PHP: Platform.COMPOSER.value,
            Language.PYTHON: Platform.PIP.value,
            Language.R: Platform.CRAN.value,
            Language.RUBY: Platform.GEM.value,
            Language.RUST: Platform.CARGO.value,
            Language.SWIFT: Platform.SWIFT.value,
            Language.GITHUB_ACTIONS: Platform.GITHUB_ACTIONS.value,
            Language.UNKNOWN_LANGUAGE: None,
        }
        return language_to_platform.get(self)


class PackageType(Enum):
    UnknownPkg = "UnknownPackage"
    AlpmPkg = "alpm"
    ApkPkg = "apk"
    BinaryPkg = "binary"
    CocoapodsPkg = "pod"
    ConanPkg = "conan"
    DartPubPkg = "dart-pub"
    DebPkg = "deb"
    DotnetPkg = "dotnet"
    ErlangOTPPkg = "erlang-otp"
    GemPkg = "gem"
    GithubActionPkg = "github-action"
    GithubActionWorkflowPkg = "github-action-workflow"
    GoModulePkg = "go-module"
    GraalVMNativeImagePkg = "graalvm-native-image"
    HackagePkg = "hackage"
    HexPkg = "hex"
    JavaPkg = "java-archive"
    JenkinsPluginPkg = "jenkins-plugin"
    KbPkg = "msrc-kb"
    LinuxKernelPkg = "linux-kernel"
    LinuxKernelModulePkg = "linux-kernel-module"
    NixPkg = "nix"
    NpmPkg = "npm"
    PhpComposerPkg = "php-composer"
    PhpPeclPkg = "php-pecl-pkg"
    PortagePkg = "portage"
    PythonPkg = "python"
    Rpkg = "R-package"
    RpmPkg = "rpm"
    RustPkg = "rust-crate"
    SwiftPkg = "swift"
    WordpressPluginPkg = "wordpress-plugin"

    def get_platform_value(self) -> str | None:
        package_type_to_platform = {
            PackageType.CocoapodsPkg: Platform.SWIFT.value,
            PackageType.ConanPkg: Platform.CONAN.value,
            PackageType.DartPubPkg: Platform.PUB.value,
            PackageType.DotnetPkg: Platform.NUGET.value,
            PackageType.ErlangOTPPkg: Platform.ERLANG.value,
            PackageType.GemPkg: Platform.GEM.value,
            PackageType.GithubActionPkg: Platform.GITHUB_ACTIONS.value,
            PackageType.GithubActionWorkflowPkg: Platform.GITHUB_ACTIONS.value,
            PackageType.GoModulePkg: Platform.GO.value,
            PackageType.HackagePkg: Platform.CABAL.value,
            PackageType.HexPkg: Platform.ERLANG.value,
            PackageType.JavaPkg: Platform.MAVEN.value,
            PackageType.NpmPkg: Platform.NPM.value,
            PackageType.PhpComposerPkg: Platform.COMPOSER.value,
            PackageType.PhpPeclPkg: Platform.COMPOSER.value,
            PackageType.PythonPkg: Platform.PIP.value,
            PackageType.Rpkg: Platform.CRAN.value,
            PackageType.RustPkg: Platform.CARGO.value,
            PackageType.SwiftPkg: Platform.SWIFT.value,
            PackageType.AlpmPkg: Platform.ALPM.value,
            PackageType.ApkPkg: Platform.APK.value,
            PackageType.BinaryPkg: Platform.BINARY.value,
            PackageType.DebPkg: Platform.DEBIAN.value,
            PackageType.GraalVMNativeImagePkg: Platform.GRAALVM_NATIVE_IMAGE.value,
            PackageType.JenkinsPluginPkg: Platform.JENKINS_PLUGIN.value,
            PackageType.KbPkg: Platform.MSRC_KB.value,
            PackageType.LinuxKernelPkg: Platform.LINUX_KERNEL.value,
            PackageType.LinuxKernelModulePkg: Platform.LINUX_KERNEL_MODULE.value,
            PackageType.NixPkg: Platform.NIXPKGS.value,
            PackageType.PortagePkg: Platform.PORTAGE.value,
            PackageType.RpmPkg: Platform.RPM.value,
            PackageType.WordpressPluginPkg: Platform.WORDPRESS_PLUGIN.value,
            PackageType.UnknownPkg: None,
        }
        return package_type_to_platform.get(self, Platform.UNKNOWN.value)


class Digest(BaseModel):
    algorithm: str | None = Field(min_length=1)
    value: str | None = Field(min_length=1)


class Artifact(BaseModel):
    url: str = Field(min_length=1)
    integrity: Digest | None = None


class HealthMetadata(BaseModel):
    latest_version: str | None = Field(default=None, min_length=1)
    latest_version_created_at: str | datetime | None = None
    artifact: Artifact | None = None
    authors: str | None = Field(default=None, min_length=1)

    @field_validator("latest_version_created_at", mode="before")
    @classmethod
    def validate_latest_version_created_at(
        cls,
        value: str | datetime | None,
    ) -> str | datetime | None:
        if isinstance(value, str) and len(value) < 1:
            error_message = (
                "latest_version_created_at must be at least 1 character long when it is a string."
            )
            raise ValueError(error_message)
        return value


class Package(BaseModel):
    name: str = Field(min_length=1)
    version: str = Field(min_length=1)
    language: Language
    licenses: list[str]
    locations: list[Location]
    type: PackageType
    advisories: list[Advisory] | None = None
    dependencies: list["Package"] | None = None
    found_by: str | None = Field(default=None, min_length=1)
    health_metadata: HealthMetadata | None = None
    is_dev: bool = False
    metadata: object | None = None
    p_url: str = Field(min_length=1)

    @property
    def id_(self) -> str:
        return self.id_by_hash()

    def id_by_hash(self) -> str:
        try:
            obj_data = {
                "name": self.name,
                "version": self.version,
                "language": self.language.value,
                "type": self.type.value,
                "p_url": self.p_url,
            }
            obj_str = json.dumps(obj_data, sort_keys=True)
            return hashlib.sha256(obj_str.encode()).hexdigest()
        except Exception as exc:  # noqa: BLE001
            return f"Could not build ID for object={self}: {exc}"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Package):
            return False

        return (
            self.name == other.name
            and self.version == other.version
            and self.language == other.language
            and self.licenses == other.licenses
            and self.locations == other.locations
            and self.type == other.type
            and self.advisories == other.advisories
            and self.dependencies == other.dependencies
            and self.found_by == other.found_by
            and self.health_metadata == other.health_metadata
            and self.is_dev == other.is_dev
            and self.metadata == other.metadata
            and self.p_url == other.p_url
        )

    def __hash__(self) -> int:
        return hash(self.id_)

    @field_validator("licenses")
    @classmethod
    def check_licenses_min_length(cls, value: list[str]) -> list[str]:
        for license_str in value:
            if len(license_str) < 1:
                error_message = "Each license string must be at least 1 character long."
                raise ValueError(error_message)
        return value



from typing import TYPE_CHECKING, cast

if TYPE_CHECKING:  # pragma: no cover
    from collections.abc import ItemsView

from labels.model.file import DependencyType, Location, LocationReadCloser, Scope
from labels.model.indexables import IndexedDict, ParsedValue
from labels.model.package import Language, Package, PackageType
from labels.model.relationship import Relationship
from labels.model.resolver import Resolver
from labels.parsers.cataloger.generic.parser import Environment
from labels.parsers.cataloger.php.package import package_url
from labels.parsers.collection.json import parse_json_with_tree_sitter

EMPTY_DICT: IndexedDict[str, ParsedValue] = IndexedDict()


def _get_location(location: Location, sourceline: int, *, is_dev: bool) -> Location:
    location.scope = Scope.DEV if is_dev else Scope.PROD
    if location.coordinates:
        c_upd = {"line": sourceline}
        l_upd = {"coordinates": location.coordinates.model_copy(update=c_upd)}
        location.dependency_type = DependencyType.DIRECT
        return location.model_copy(update=l_upd)
    return location


def _get_packages(
    reader: LocationReadCloser,
    dependencies: IndexedDict[str, ParsedValue],
    *,
    is_dev: bool,
) -> list[Package]:
    if not dependencies:
        return []

    items: ItemsView[str, ParsedValue] = dependencies.items()
    return [
        Package(
            name=name,
            version=version,
            locations=[
                _get_location(
                    reader.location,
                    dependencies.get_key_position(name).start.line,
                    is_dev=is_dev,
                ),
            ],
            language=Language.PHP,
            licenses=[],
            type=PackageType.PhpComposerPkg,
            p_url=package_url(name, version),
            is_dev=is_dev,
        )
        for name, version in items
        if isinstance(version, str)
    ]


def parse_composer_json(
    _: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    content = cast(
        "IndexedDict[str, ParsedValue]",
        parse_json_with_tree_sitter(reader.read_closer.read()),
    )
    deps: IndexedDict[str, ParsedValue] = cast(
        "IndexedDict[str, ParsedValue]",
        content.get("require", EMPTY_DICT),
    )
    dev_deps: IndexedDict[str, ParsedValue] = cast(
        "IndexedDict[str, ParsedValue]",
        content.get("require-dev", EMPTY_DICT),
    )
    packages = [
        *_get_packages(reader, deps, is_dev=False),
        *_get_packages(reader, dev_deps, is_dev=True),
    ]
    return packages, []



import logging
from copy import deepcopy

from pydantic import ValidationError

from labels.model.file import DependencyType, Location, LocationReadCloser, Scope
from labels.model.indexables import IndexedDict, ParsedValue
from labels.model.package import Language, Package, PackageType
from labels.model.relationship import Relationship
from labels.model.resolver import Resolver
from labels.parsers.cataloger.generic.parser import Environment
from labels.parsers.cataloger.python.model import PythonPackage
from labels.parsers.cataloger.python.package import package_url
from labels.parsers.collection import toml
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)


def _get_location(location: Location, sourceline: int, scope: Scope) -> Location:
    location.scope = scope
    location.dependency_type = DependencyType.DIRECT
    if location.coordinates:
        c_upd = {"line": sourceline}
        l_upd = {"coordinates": location.coordinates.model_copy(update=c_upd)}
        return location.model_copy(update=l_upd)
    return location


def _get_packages(
    reader: LocationReadCloser,
    toml_packages: IndexedDict[str, ParsedValue],
    *,
    is_dev: bool = False,
) -> list[Package]:
    result = []
    scope = Scope.DEV if is_dev else Scope.PROD
    for package, version_data in toml_packages.items():
        version: str = ""
        if isinstance(version_data, str):
            version = version_data.strip("=<>~^ ")
        if isinstance(version_data, IndexedDict):
            version = str(version_data.get("version", "*")).strip("=<>~^ ")

        if not package or not version or "*" in version:
            continue
        current_location = deepcopy(reader.location)
        try:
            result.append(
                Package(
                    name=package,
                    version=version,
                    locations=[
                        (
                            _get_location(
                                current_location,
                                package.position.start.line
                                if isinstance(package, IndexedDict)
                                else toml_packages.get_key_position(package).start.line,
                                scope,
                            )
                        ),
                    ],
                    language=Language.PYTHON,
                    type=PackageType.PythonPkg,
                    metadata=PythonPackage(
                        name=package,
                        version=version,
                    ),
                    p_url=package_url(
                        name=package,
                        version=version,
                        package=None,
                    ),
                    licenses=[],
                    is_dev=is_dev,
                ),
            )
        except ValidationError as ex:
            LOGGER.warning(
                "Malformed package. Required fields are missing or data types are incorrect.",
                extra={
                    "extra": {
                        "exception": format_exception(str(ex)),
                        "location": reader.location.path(),
                    },
                },
            )
            continue
    return result


def parse_pipfile_deps(
    _resolver: Resolver | None,
    _env: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    packages = []
    file_content = reader.read_closer.read()
    toml_content: IndexedDict[str, ParsedValue] = toml.parse_toml_with_tree_sitter(file_content)
    toml_packages = toml_content.get("packages")
    if not isinstance(toml_packages, IndexedDict):
        return [], []
    packages = _get_packages(reader, toml_packages)
    dev_deps = toml_content.get("dev-packages")
    if isinstance(dev_deps, IndexedDict):
        dev_pkgs = _get_packages(reader, dev_deps, is_dev=True)
        packages.extend(dev_pkgs)
    return packages, []



import logging
from copy import deepcopy
from typing import cast

from packageurl import PackageURL
from pydantic import ValidationError

from labels.model.file import DependencyType, Location, Scope
from labels.model.indexables import IndexedDict, IndexedList, ParsedValue
from labels.model.package import Language, Package, PackageType
from labels.parsers.cataloger.php.model import (
    PhpComposerAuthors,
    PhpComposerExternalReference,
    PhpComposerInstalledEntry,
)
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)

EMPTY_LIST: IndexedList[str] = IndexedList()


def package_url(name: str, version: str) -> str:
    if not name:
        error_msg = "Package name cannot be empty"
        raise ValueError(error_msg)

    fields = name.split("/")

    vendor = ""
    if len(fields) == 1:
        name = fields[0]
    else:
        vendor = fields[0]
        name = "-".join(fields[1:])

    return PackageURL(  # type: ignore
        type="composer",
        namespace=vendor,
        name=name,
        version=version,
        qualifiers=None,
        subpath="",
    ).to_string()


def new_package_from_composer(
    package: IndexedDict[str, ParsedValue],
    location: Location,
    *,
    is_dev: bool = False,
) -> Package | None:
    empty_list_dict: IndexedList[IndexedDict[str, str]] = IndexedList()
    new_location = deepcopy(location)
    new_location.scope = Scope.DEV if is_dev else Scope.PROD

    try:
        source = cast(IndexedDict[str, str], package.get("source"))
        dist = cast(IndexedDict[str, str], package.get("dist"))
        name = cast(str, package.get("name"))
        version = cast(str, package.get("version"))
        if not name or not version:
            return None

        if new_location.coordinates:
            new_location.dependency_type = DependencyType.DIRECT
            new_location.coordinates.line = package.get_key_position("name").start.line

        return Package(
            name=name,
            version=version,
            locations=[new_location],
            language=Language.PHP,
            licenses=list(cast(IndexedList[str], package.get("license", EMPTY_LIST))),
            type=PackageType.PhpComposerPkg,
            p_url=package_url(name, version),
            metadata=PhpComposerInstalledEntry(
                name=name,
                version=version,
                source=PhpComposerExternalReference(
                    type=source.get("type") or None,
                    url=source.get("url") or None,
                    reference=source.get("reference") or None,
                    shasum=source.get("shasum") or None,
                )
                if source
                else None,
                dist=PhpComposerExternalReference(
                    type=dist.get("type") or None,
                    url=dist.get("url") or None,
                    reference=dist.get("reference") or None,
                    shasum=dist.get("shasum") or None,
                )
                if dist
                else None,
                require=cast(dict[str, str], package.get("require"))
                if isinstance(package.get("require"), IndexedDict)
                else None,
                provide=cast(dict[str, str], package.get("provide"))
                if isinstance(package.get("provide"), IndexedDict)
                else None,
                require_dev=cast(dict[str, str], package.get("require-dev"))
                if isinstance(package.get("require-dev"), IndexedDict)
                else None,
                suggest=cast(dict[str, str], package.get("suggest"))
                if isinstance(package.get("suggest"), IndexedDict)
                else None,
                license=cast(list[str], package.get("license"))
                if isinstance(package.get("license"), IndexedList)
                else None,
                type=cast(str, package.get("type"))
                if isinstance(package.get("type"), str)
                else None,
                notification_url=cast(str, package.get("notification-url"))
                if isinstance(package.get("notification-url"), str)
                else None,
                bin=cast(list[str], package.get("bin"))
                if isinstance(package.get("bin"), IndexedList)
                else None,
                authors=[
                    PhpComposerAuthors(
                        name=cast(str, x.get("name")),
                        email=x.get("email"),
                        homepage=x.get("homepage"),
                    )
                    for x in cast(
                        list[IndexedDict[str, str]],
                        package.get("authors", empty_list_dict),
                    )
                ],
                description=cast(str, package.get("description"))
                if isinstance(package.get("description"), str)
                else None,
                homepage=cast(str, package.get("homepage"))
                if isinstance(package.get("homepage"), str)
                else None,
                keywords=cast(list[str], package.get("keywords")),
                time=cast(str, package.get("time"))
                if isinstance(package.get("time"), str)
                else None,
            ),
            is_dev=is_dev,
        )
    except ValidationError as ex:
        LOGGER.warning(
            "Malformed package. Required fields are missing or data types are incorrect.",
            extra={
                "extra": {
                    "exception": format_exception(str(ex)),
                    "location": new_location.path(),
                },
            },
        )
        return None


def package_url_from_pecl(pkg_name: str, version: str) -> str:
    return PackageURL(  # type: ignore
        type="pecl",
        namespace="",
        name=pkg_name,
        version=version,
        qualifiers=None,
        subpath="",
    ).to_string()



from copy import deepcopy
from typing import cast

from labels.model.file import LocationReadCloser
from labels.model.indexables import IndexedDict, IndexedList
from labels.model.package import Package
from labels.model.relationship import Relationship
from labels.model.resolver import Resolver
from labels.parsers.cataloger.cpp.package import new_conan_lock_dep
from labels.parsers.cataloger.generic.parser import Environment
from labels.parsers.collection.json import parse_json_with_tree_sitter


def parse_conan_lock(
    _: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    packages = []
    conan_file = cast(
        IndexedDict[str, IndexedList[str]],
        parse_json_with_tree_sitter(reader.read_closer.read()),
    )
    for index, dep_line in enumerate(conan_file.get("requires", [])):
        location = deepcopy(reader.location)
        if location.coordinates:
            location.coordinates.line = (
                conan_file["requires"]
                .get_position(
                    index,
                )
                .start.line
            )
        packages.append(
            new_conan_lock_dep(dep_line, location),
        )
    for index, dep_line in enumerate(conan_file.get("build_requires", [])):
        location = deepcopy(reader.location)
        if location.coordinates:
            location.coordinates.line = conan_file["build_requires"].get_position(index).start.line
        packages.append(
            new_conan_lock_dep(dep_line, location, is_dev=True),
        )
    return packages, []



import logging
from collections.abc import Callable
from fnmatch import fnmatch

import reactivex
from pydantic import BaseModel, ConfigDict
from reactivex import Observable
from reactivex.abc import ObserverBase, SchedulerBase
from reactivex.scheduler import ThreadPoolScheduler

from labels.model.file import Location, LocationReadCloser
from labels.model.package import Package
from labels.model.relationship import Relationship
from labels.model.resolver import Resolver
from labels.parsers.cataloger.alpine.parse_apk_db import parse_apk_db
from labels.parsers.cataloger.arch.parse_alpm import parse_alpm_db
from labels.parsers.cataloger.debian.parse_dpkg_db import parse_dpkg_db
from labels.parsers.cataloger.generic.parser import Environment, Parser
from labels.parsers.cataloger.redhat.parse_rpm_db import parse_rpm_db

LOGGER = logging.getLogger(__name__)


class Request(BaseModel):
    real_path: str
    parser: Parser
    parser_name: str
    model_config = ConfigDict(frozen=True)


class Task(BaseModel):
    location: Location
    parser: Parser
    parser_name: str
    model_config = ConfigDict(frozen=True)


def execute_parsers(
    resolver: Resolver,
    environment: Environment,
) -> Callable[[Observable[Task]], Observable]:
    def _handle(source: Observable[Task]) -> Observable:
        def subscribe(
            observer: ObserverBase[tuple[list[Package], list[Relationship]]],
            scheduler: ThreadPoolScheduler | None = None,
        ) -> reactivex.abc.DisposableBase:
            def on_next(value: Task) -> None:
                LOGGER.debug("Working on %s", value.location.access_path)
                content_reader = resolver.file_contents_by_location(value.location)
                try:
                    if content_reader is not None and (
                        result := value.parser(
                            resolver,
                            environment,
                            LocationReadCloser(
                                location=value.location,
                                read_closer=content_reader,
                            ),
                        )
                    ):
                        discover_packages, relationships = result
                        for pkg in discover_packages:
                            pkg.found_by = value.parser_name
                        observer.on_next((discover_packages, relationships))
                except Exception as ex:  # noqa: BLE001
                    observer.on_error(ex)

            return source.subscribe(
                on_next,
                observer.on_error,
                observer.on_completed,
                scheduler=scheduler,
            )

        return reactivex.create(subscribe)  # type: ignore

    return _handle


def on_next_db_file(
    source: reactivex.Observable[str],
) -> reactivex.Observable[Request]:
    def subscribe(
        observer: ObserverBase[Request],
        scheduler: SchedulerBase | None = None,
    ) -> reactivex.abc.DisposableBase:
        def on_next(value: str) -> None:
            try:
                if "lib/apk/db/installed" in value:
                    observer.on_next(
                        Request(
                            real_path=value,
                            parser=parse_apk_db,
                            parser_name="apk-db-selector",
                        ),
                    )
                elif any(
                    fnmatch(value, x)
                    for x in (
                        "**/var/lib/dpkg/status",
                        "*var/lib/dpkg/status",
                        "/var/lib/dpkg/status",
                        "**/var/lib/dpkg/status.d/*",
                        "*var/lib/dpkg/status.d/*",
                        "/var/lib/dpkg/status.d/*",
                        "**/lib/opkg/info/*.control",
                        "*lib/opkg/info/*.control",
                        "/lib/opkg/info/*.control",
                        "**/lib/opkg/status",
                        "*lib/opkg/status",
                        "/lib/opkg/status",
                    )
                ):
                    observer.on_next(
                        Request(
                            real_path=value,
                            parser=parse_dpkg_db,
                            parser_name="dpkg-db-selector",
                        ),
                    )
                elif any(
                    fnmatch(value, pattern)
                    for pattern in (
                        "**/var/lib/pacman/local/**/desc",
                        "var/lib/pacman/local/**/desc",
                        "/var/lib/pacman/local/**/desc",
                    )
                ):
                    observer.on_next(
                        Request(
                            real_path=value,
                            parser=parse_alpm_db,
                            parser_name="alpm-db-selector",
                        ),
                    )
                elif any(
                    fnmatch(value, x)
                    for x in (
                        (
                            "**/{var/lib,usr/share,usr/lib/sysimage}"
                            "/rpm/{Packages,Packages.db,rpmdb.sqlite}"
                        ),
                        (
                            "/{var/lib,usr/share,usr/lib/sysimage}"
                            "/rpm/{Packages,Packages.db,rpmdb.sqlite}"
                        ),
                        "**/rpmdb.sqlite",
                        "**/var/lib/rpm/Packages",
                        "**/var/lib/rpm/Packages.db",
                        "**/var/lib/rpm/rpmdb.sqlite",
                        "**/usr/share/rpm/Packages",
                        "**/usr/share/rpm/Packages.db",
                        "**/usr/share/rpm/rpmdb.sqlite",
                        "**/usr/lib/sysimage/rpm/Packages",
                        "**/usr/lib/sysimage/rpm/Packages.db",
                        "**/usr/lib/sysimage/rpm/rpmdb.sqlite",
                        "/var/lib/rpm/Packages",
                        "/var/lib/rpm/Packages.db",
                        "/var/lib/rpm/rpmdb.sqlite",
                        "/usr/share/rpm/Packages",
                        "/usr/share/rpm/Packages.db",
                        "/usr/share/rpm/rpmdb.sqlite",
                        "/usr/lib/sysimage/rpm/Packages",
                        "/usr/lib/sysimage/rpm/Packages.db",
                        "/usr/lib/sysimage/rpm/rpmdb.sqlite",
                    )
                ):
                    observer.on_next(
                        Request(
                            real_path=value,
                            parser=parse_rpm_db,
                            parser_name="environment-parser",
                        ),
                    )
            except Exception as ex:  # noqa: BLE001
                observer.on_error(ex)

        return source.subscribe(
            on_next,
            observer.on_error,
            observer.on_completed,
            scheduler=scheduler,
        )

    return reactivex.create(subscribe)



import os
from collections.abc import Callable
from os.path import split

from labels.advisories.match_versions import (
    convert_semver_to_range,
    is_single_version,
    match_version_ranges,
)
from labels.model.core import OutputFormat, SbomConfig
from labels.model.file import Location
from labels.model.package import Package
from labels.model.relationship import Relationship
from labels.model.resolver import Resolver
from labels.output.cyclonedx.output_handler import format_cyclonedx_sbom
from labels.output.fluid.output_handler import format_fluid_sbom
from labels.output.spdx.output_handler import format_spdx_sbom

_FORMAT_HANDLERS: dict[OutputFormat, Callable] = {
    OutputFormat.FLUID_JSON: format_fluid_sbom,
    OutputFormat.CYCLONEDX_JSON: format_cyclonedx_sbom,
    OutputFormat.CYCLONEDX_XML: format_cyclonedx_sbom,
    OutputFormat.SPDX_JSON: format_spdx_sbom,
    OutputFormat.SPDX_XML: format_spdx_sbom,
}

POSSIBLE_LOCK_FILES = {
    "Pipfile": ["Pipfile.lock"],
    "Cargo.toml": ["Cargo.lock"],
    "conanfile.txt": ["conan.lock"],
    "conanfile.py": ["conan.lock"],
    "composer.json": ["composer.lock"],
    "mix.exs": ["mix.lock"],
    "packages.config": ["packages.lock.json"],
    "package.json": [
        "package-lock.json",
        "yarn.lock",
        "pnpm-lock.yaml",
        # For angular libraries
        "ng-package.json",
    ],
    "Gemfile": ["Gemfile.lock", "gems.locked"],
    "pyproject.toml": ["poetry.lock"],
    "package.swift": ["Package.resolved"],
    "Package.swift": ["Package.resolved"],
}


def extract_file_info(path: str) -> str:
    _, file_info = split(path)
    return file_info


def update_package_locations(packages: list[Package]) -> None:
    for package in packages:
        lock_files, non_lock_files = _find_lock_and_non_lock_files(package.locations)

        for lock_file in lock_files:
            for non_lock_file in non_lock_files:
                if same_directory(non_lock_file.access_path, lock_file.access_path):
                    lock_file.dependency_type = non_lock_file.dependency_type
                    lock_file.scope = non_lock_file.scope


def _find_lock_and_non_lock_files(
    locations: list[Location],
) -> tuple[list[Location], list[Location]]:
    non_lock_files = []
    lock_files = []

    for location in locations:
        file_info = extract_file_info(location.path())
        if file_info in POSSIBLE_LOCK_FILES:
            non_lock_files.append(location)
        elif any(file_info == lock for locks in POSSIBLE_LOCK_FILES.values() for lock in locks):
            lock_files.append(location)

    return lock_files, non_lock_files


def split_version_range(v1: str, v2: str) -> tuple[str, str] | None:
    v1_single = is_single_version(convert_semver_to_range(v1))
    v2_single = is_single_version(convert_semver_to_range(v2))
    if not v1_single and v2_single:
        return v1, v2

    if not v2_single and v1_single:
        return v2, v1

    return None


def parent_dir(path: str) -> str:
    p = path.rstrip(os.sep)
    if os.sep not in p:
        return ""
    return p.rsplit(os.sep, 1)[0]


def same_directory(path1: str | None, path2: str | None) -> bool:
    if path1 is None or path2 is None:
        return False

    if path1.rsplit(os.sep, 1)[-1] in POSSIBLE_LOCK_FILES.get(
        path2.rsplit(os.sep, 1)[-1],
        [],
    ) or path2.rsplit(os.sep, 1)[-1] in POSSIBLE_LOCK_FILES.get(path1.rsplit(os.sep, 1)[-1], []):
        return parent_dir(path1) == parent_dir(path2)

    return False


def _merge_locations(target: Package, source: Package) -> None:
    for loc in source.locations:
        if loc not in target.locations:
            target.locations.append(loc)


def _find_merge_target(pkg: Package, candidates: list[Package]) -> Package | None:
    for existing in candidates:
        if existing.name != pkg.name:
            continue
        if _share_parent_dir(existing.locations, pkg.locations) and match_version_ranges(
            pkg.version,
            existing.version,
        ):
            return existing
    return None


def _share_parent_dir(locations_a: list[Location], locations_b: list[Location]) -> bool:
    return any(
        same_directory(a.access_path, b.access_path) for a in locations_a for b in locations_b
    )


def merge_packages(packages: list[Package]) -> list[Package]:
    merged: dict[str, Package] = {}
    for pkg in packages:
        if pkg.id_ in merged:
            _merge_locations(merged[pkg.id_], pkg)
            continue

        target = _find_merge_target(pkg, list(merged.values()))
        if target:
            split = split_version_range(target.version, pkg.version)
            if split:
                _, fixed_ver = split
                if fixed_ver == pkg.version:
                    _merge_locations(pkg, target)
                    del merged[target.id_]
                    merged[pkg.id_] = pkg
                    continue
            _merge_locations(target, pkg)
        else:
            merged[pkg.id_] = pkg

    update_package_locations(packages)
    return list(merged.values())


def dispatch_sbom_output(
    *,
    packages: list[Package],
    relationships: list[Relationship],
    config: SbomConfig,
    resolver: Resolver,
) -> None:
    processed_packages = merge_packages(packages)

    handler = _FORMAT_HANDLERS[config.output_format]
    handler(
        packages=processed_packages,
        relationships=relationships,
        config=config,
        resolver=resolver,
    )



import re
from collections.abc import Iterable
from copy import deepcopy

from gemfileparser import Dependency, GemfileParser

from labels.model.file import DependencyType, Location, LocationReadCloser, Scope
from labels.model.package import Language, Package, PackageType
from labels.model.relationship import Relationship
from labels.model.resolver import Resolver
from labels.parsers.cataloger.generic.parser import Environment

from .package import package_url

GEMFILE_DEP: re.Pattern[str] = re.compile(
    r'^\s*(?P<gem>gem ".*?",?( "[><~=]{0,2}\s?[\d\.]+",?){0,2})',
)
NOT_PROD_DEP: re.Pattern[str] = re.compile(
    r":group => \[?[:\w\-, ]*(:development|:test)",
)
NOT_PROD_GROUP: re.Pattern[str] = re.compile(r"(\s*)group :(test|development)")
GEM_LOCK_DEP: re.Pattern[str] = re.compile(
    r"^\s{4}(?P<gem>[^\s]*)\s\([^\d]*(?P<version>.*)\)$",
)


def _is_dev_dependency(line: str) -> bool:
    return bool(NOT_PROD_DEP.search(line))


def format_requirements(requirements: list[str]) -> str:
    formatted: str = ""
    if len(requirements) == 0:
        return formatted
    requirements = [req.replace(" ", "") for req in requirements]
    first_req = requirements[0]
    if "~" in first_req:
        if len(requirements) == 1:
            dot_times = first_req.count(".")
            if dot_times <= 1:
                formatted = first_req.replace("~>", "^")
            else:
                formatted = first_req.replace("~>", "~")
        elif len(requirements) == 2:
            sec_req = requirements[1]
            if ">=" in sec_req:
                formatted = sec_req.replace(">=", "^")
            elif "!=" in sec_req:
                sec_ver = sec_req.replace("!=", "")
                first_ver = first_req.replace("!=", "")
                formatted = f">={first_ver} <{sec_ver}  || >{sec_ver}"
    else:
        formatted = " ".join(requirements)
    return formatted


def match_dep_criteria(
    column_list: Iterable[str],
) -> Dependency:
    dep = Dependency()
    for column in column_list:
        for criteria, criteria_regex in GemfileParser.gemfile_regexes.items():
            match = criteria_regex.match(column)
            if match:
                if criteria == "requirement":
                    dep.requirement.append(match.group(criteria))
                else:
                    setattr(dep, criteria, match.group(criteria))
                break
    return dep


def parse_line(in_line: str, *, gem_file: bool) -> tuple[str, str]:
    line: list[str] = []
    if gem_file:
        line = in_line.split(",")
    else:
        line = in_line.split(" ", maxsplit=1)
        if len(line) > 1 and (", " in line[1]):
            line = [line[0], *line[1].split(", ")]

    column_list: list[str] = []
    for column in line:
        stripped_column = (
            column.replace("'", "")
            .replace('"', "")
            .replace("%q<", "")
            .replace("(", "")
            .replace(")", "")
            .replace("[", "")
            .replace("]", "")
            .strip()
        )
        column_list.append(stripped_column)

    deps = match_dep_criteria(column_list)
    deps_dict = deps.to_dict()
    product: str = deps_dict["name"]
    version: str = format_requirements(deps_dict["requirement"])
    return product, version


def create_dependency(
    line_number: int,
    location: Location,
    matched: re.Match[str],
    *,
    is_dev: bool = False,
) -> Package | None:
    gem_info = GemfileParser.preprocess(matched.group("gem"))[3:]
    product, version = parse_line(gem_info, gem_file=True)
    new_location = deepcopy(location)
    new_location.scope = Scope.DEV if is_dev else Scope.PROD
    if new_location.coordinates:
        new_location.coordinates.line = line_number
        new_location.dependency_type = DependencyType.DIRECT
    if not version:
        return None
    return Package(
        name=product,
        version=version,
        type=PackageType.GemPkg,
        locations=[new_location],
        p_url=package_url(product, version),
        metadata=None,
        language=Language.RUBY,
        licenses=[],
        is_dev=is_dev,
    )


def parse_gemfile(
    _: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    packages = []
    in_group_block = False
    end_line = ""

    for line_number, line in enumerate(
        reader.read_closer.read().splitlines(),
        1,
    ):
        if in_group_block:
            if line == end_line:
                in_group_block = False
            elif (
                (matched := GEMFILE_DEP.search(line))
                and matched
                and (
                    dependency := create_dependency(
                        line_number,
                        reader.location,
                        matched,
                        is_dev=_is_dev_dependency(line),
                    )
                )
            ):
                packages.append(dependency)
            continue

        if match_group := NOT_PROD_GROUP.search(line):
            in_group_block = True
            end_line = f"{match_group.group(1)}end"
            continue

        if (
            (matched := GEMFILE_DEP.search(line))
            and matched
            and (
                dependency := create_dependency(
                    line_number,
                    reader.location,
                    matched,
                    is_dev=_is_dev_dependency(line),
                )
            )
        ):
            packages.append(dependency)

    return packages, []



import base64
from contextlib import suppress

from labels.enrichers.javascript.get import NPMPackage, NPMPackageLicense, get_npm_package
from labels.enrichers.utils import infer_algorithm
from labels.model.package import Artifact, Digest, HealthMetadata, Package
from labels.utils.licenses.validation import validate_licenses


def _get_author(npm_package: NPMPackage) -> str | None:
    author: str | None = None
    if "author" in npm_package:
        package_author = npm_package["author"]
        if isinstance(package_author, dict) and "name" in package_author:
            author = package_author["name"]
            if "email" in package_author:
                author = f"{author} <{package_author['email']}>"
        elif package_author and isinstance(package_author, str):
            author = str(package_author)
        return author
    return None


def _get_latest_version_info(
    npm_package: NPMPackage,
) -> tuple[str | None, str | None]:
    latest_version = None
    latest_version_created_at = None

    if npm_package.get("dist-tags"):
        latest_version = npm_package["dist-tags"]["latest"]
        latest_version_created_at = npm_package["time"][latest_version]

    return latest_version, latest_version_created_at


def _get_artifact_info(
    npm_package: NPMPackage,
    current_version: str,
) -> Artifact | None:
    current_package = npm_package["versions"].get(current_version)
    artifact = None

    if current_package:
        with suppress(KeyError):
            digest_value = current_package.get("dist", {}).get("integrity") or None

            if digest_value:
                algorithm, digest_hash = digest_value.split("-", 1)
                if algorithm == "sha512":
                    binary_hash = base64.b64decode(digest_hash)
                    digest_hash = binary_hash.hex()

                artifact = Artifact(
                    url=current_package["dist"]["tarball"],
                    integrity=Digest(
                        algorithm=infer_algorithm(digest_hash),
                        value=digest_hash,
                    ),
                )

    return artifact


def _set_health_metadata(package: Package, npm_package: NPMPackage) -> None:
    latest_version, latest_version_created_at = _get_latest_version_info(
        npm_package,
    )
    package.health_metadata = HealthMetadata(
        latest_version=latest_version,
        latest_version_created_at=latest_version_created_at,
        artifact=_get_artifact_info(npm_package, package.version),
        authors=_get_author(npm_package),
    )


def complete_package(package: Package) -> Package:
    npm_package = get_npm_package(package.name)
    if not npm_package:
        return package

    _set_health_metadata(package, npm_package)

    licenses = npm_package.get("license")
    if licenses and isinstance(licenses, (str | list | dict)):
        package.licenses = validate_licenses(handle_licenses(licenses))

    return package


def handle_licenses(
    licenses: str | list[str | dict[str, str]] | NPMPackageLicense,
) -> list[str]:
    if isinstance(licenses, dict):
        return [licenses["type"]] if "type" in licenses else []
    if isinstance(licenses, list):
        licenses_list = []
        for license_item in licenses:
            if isinstance(license_item, str):
                licenses_list.append(license_item)
            if isinstance(license_item, dict) and license_item["type"]:
                licenses_list.append(license_item["type"])
        return licenses_list
    return [licenses]



import logging
import os
from collections.abc import Iterator
from ctypes import c_void_p, cdll
from itertools import count

import networkx as nx
import tree_sitter_python
from tree_sitter import Language, Node, Parser, Tree

from labels.config.utils import TREE_SITTER_PARSERS

LOGGER = logging.getLogger(__name__)

PYTHON_FIELDS = {
    "aliased_import": [
        "alias",
        "name",
    ],
    "as_pattern": [
        "alias",
    ],
    "assignment": [
        "left",
        "right",
        "type",
    ],
    "attribute": [
        "attribute",
        "object",
    ],
    "augmented_assignment": [
        "left",
        "operator",
        "right",
    ],
    "binary_operator": [
        "left",
        "operator",
        "right",
    ],
    "block": [
        "alternative",
    ],
    "boolean_operator": [
        "left",
        "operator",
        "right",
    ],
    "call": [
        "arguments",
        "function",
    ],
    "case_clause": [
        "consequence",
        "guard",
    ],
    "class_definition": [
        "body",
        "name",
        "superclasses",
        "type_parameters",
    ],
    "comparison_operator": [
        "operators",
    ],
    "decorated_definition": [
        "definition",
    ],
    "default_parameter": [
        "name",
        "value",
    ],
    "dict_pattern": [
        "key",
        "value",
    ],
    "dictionary_comprehension": [
        "body",
    ],
    "elif_clause": [
        "condition",
        "consequence",
    ],
    "else_clause": [
        "body",
    ],
    "except_clause": [
        "alias",
        "value",
    ],
    "exec_statement": [
        "code",
    ],
    "for_in_clause": [
        "left",
        "right",
    ],
    "for_statement": [
        "alternative",
        "body",
        "left",
        "right",
    ],
    "format_expression": [
        "expression",
        "format_specifier",
        "type_conversion",
    ],
    "function_definition": [
        "body",
        "name",
        "parameters",
        "return_type",
        "type_parameters",
    ],
    "future_import_statement": [
        "name",
    ],
    "generator_expression": [
        "body",
    ],
    "if_statement": [
        "alternative",
        "condition",
        "consequence",
    ],
    "import_from_statement": [
        "module_name",
        "name",
    ],
    "import_statement": [
        "name",
    ],
    "interpolation": [
        "expression",
        "format_specifier",
        "type_conversion",
    ],
    "keyword_argument": [
        "name",
        "value",
    ],
    "lambda": [
        "body",
        "parameters",
    ],
    "list_comprehension": [
        "body",
    ],
    "match_statement": [
        "body",
        "subject",
    ],
    "named_expression": [
        "name",
        "value",
    ],
    "not_operator": [
        "argument",
    ],
    "pair": [
        "key",
        "value",
    ],
    "print_statement": [
        "argument",
    ],
    "raise_statement": [
        "cause",
    ],
    "set_comprehension": [
        "body",
    ],
    "subscript": [
        "subscript",
        "value",
    ],
    "try_statement": [
        "body",
    ],
    "type_alias_statement": [
        "left",
        "right",
    ],
    "typed_default_parameter": [
        "name",
        "type",
        "value",
    ],
    "typed_parameter": [
        "type",
    ],
    "unary_operator": [
        "argument",
        "operator",
    ],
    "while_statement": [
        "alternative",
        "body",
        "condition",
    ],
    "with_item": [
        "value",
    ],
    "with_statement": [
        "body",
    ],
}

SWIFT_FIELDS: dict[str, list[str]] = {
    "additive_expression": ["lhs", "op", "rhs"],
    "array_literal": ["element"],
    "array_type": ["element", "name"],
    "as_expression": ["expr", "name", "type"],
    "assignment": ["operator", "result", "target"],
    "associatedtype_declaration": ["default_value", "must_inherit", "name"],
    "await_expression": ["expr"],
    "bitwise_operation": ["lhs", "op", "rhs"],
    "call_suffix": ["name"],
    "capture_list_item": ["name", "value"],
    "catch_block": ["error"],
    "check_expression": ["name", "op", "target", "type"],
    "class_declaration": ["body", "declaration_kind", "name"],
    "comparison_expression": ["lhs", "op", "rhs"],
    "conjunction_expression": ["lhs", "op", "rhs"],
    "constructor_expression": ["constructed_type"],
    "constructor_suffix": ["name"],
    "control_transfer_statement": ["result"],
    "deinit_declaration": ["body"],
    "dictionary_literal": ["key", "value"],
    "dictionary_type": ["key", "name", "value"],
    "disjunction_expression": ["lhs", "op", "rhs"],
    "enum_entry": ["data_contents", "name", "raw_value"],
    "enum_type_parameters": ["name"],
    "equality_constraint": ["constrained_type", "must_equal", "name"],
    "equality_expression": ["lhs", "op", "rhs"],
    "for_statement": ["collection", "item"],
    "function_declaration": ["body", "default_value", "name", "return_type"],
    "function_type": ["name", "params", "return_type"],
    "guard_statement": ["bound_identifier", "condition", "name"],
    "if_statement": ["bound_identifier", "condition", "name"],
    "infix_expression": ["lhs", "op", "rhs"],
    "inheritance_constraint": ["constrained_type", "inherits_from", "name"],
    "inheritance_specifier": ["inherits_from"],
    "init_declaration": ["body", "default_value", "name"],
    "interpolated_expression": ["name", "reference_specifier", "value"],
    "lambda_function_type": ["name", "return_type"],
    "lambda_literal": ["captures", "type"],
    "lambda_parameter": ["external_name", "name", "type"],
    "line_string_literal": ["interpolation", "text"],
    "macro_declaration": ["default_value", "definition"],
    "macro_definition": ["body"],
    "multi_line_string_literal": ["interpolation", "text"],
    "multiplicative_expression": ["lhs", "op", "rhs"],
    "navigation_expression": ["element", "suffix", "target"],
    "navigation_suffix": ["suffix"],
    "nil_coalescing_expression": ["if_nil", "value"],
    "open_end_range_expression": ["start"],
    "open_start_range_expression": ["end"],
    "optional_type": ["wrapped"],
    "parameter": ["external_name", "name", "type"],
    "pattern": ["bound_identifier", "name"],
    "postfix_expression": ["operation", "target"],
    "prefix_expression": ["operation", "target"],
    "property_declaration": ["computed_value", "name", "value"],
    "protocol_body": ["body"],
    "protocol_declaration": ["body", "declaration_kind", "name"],
    "protocol_function_declaration": ["default_value", "name", "return_type"],
    "protocol_property_declaration": ["name"],
    "range_expression": ["end", "op", "start"],
    "raw_str_interpolation": ["interpolation"],
    "raw_string_literal": ["interpolation", "text"],
    "repeat_while_statement": ["bound_identifier", "condition", "name"],
    "subscript_declaration": ["default_value", "name", "return_type"],
    "switch_statement": ["expr"],
    "ternary_expression": ["condition", "if_false", "if_true"],
    "try_expression": ["expr"],
    "tuple_expression": ["name", "value"],
    "tuple_type": ["element"],
    "tuple_type_item": ["element", "name", "type"],
    "type_annotation": ["name", "type"],
    "type_arguments": ["name"],
    "type_parameter": ["name"],
    "typealias_declaration": ["name", "value"],
    "value_argument": ["name", "reference_specifier", "value"],
    "value_binding_pattern": ["mutability"],
    "while_statement": ["bound_identifier", "condition", "name"],
}


class Graph(nx.DiGraph):
    pass


NId = str
NAttrs = dict[str, str | int]


class ParsingError(Exception):
    pass


def lang_from_so(language: str) -> Language | None:
    if not TREE_SITTER_PARSERS:
        return None
    so_library_path = os.path.join(TREE_SITTER_PARSERS, f"{language}.so")
    lib = cdll.LoadLibrary(os.fspath(so_library_path))
    language_function = getattr(lib, f"tree_sitter_{language}")
    language_function.restype = c_void_p
    language_ptr = language_function()
    if not language_ptr:
        msg = f"Failed to load language: {language}"
        raise ValueError(msg)
    return Language(language_ptr)


PARSER_LANGUAGES: dict[str, Language | None] = {
    "python": Language(tree_sitter_python.language()),
    "swift": lang_from_so("swift"),
}


LANGUAGE_FIELDS: dict[str, dict[str, list[str]]] = {
    "python": PYTHON_FIELDS,
    "swift": SWIFT_FIELDS,
}


def hash_node(node: Node) -> int:
    return hash((node.end_point, node.start_point, node.type))


def _build_ast_graph(
    *,
    content: bytes,
    node: Node,
    counter: Iterator[str],
    graph: Graph,
    language: str,
    _edge_index: str | None = None,
    _parent: str | None = None,
    _parent_fields: dict[int, str] | None = None,
) -> Graph:
    if not isinstance(node, Node) or node.has_error:
        raise ParsingError

    n_id = next(counter)
    raw_l, raw_c = node.start_point

    graph.add_node(n_id, label_l=raw_l + 1, label_c=raw_c + 1, label_type=node.type)

    if _parent is not None:
        graph.add_edge(_parent, n_id, label_ast="AST", label_index=_edge_index)
        if field := (_parent_fields or {}).get(hash_node(node)):
            graph.nodes[_parent][f"label_field_{field}"] = n_id

    if not node.children:
        # Extract the text from it
        node_content = content[node.start_byte : node.end_byte]
        graph.nodes[n_id]["label_text"] = node_content.decode("latin-1")

    if node.children:
        # It's not a final node, recurse
        for edge_index, child in enumerate(node.children):
            _build_ast_graph(
                content=content,
                node=child,
                counter=counter,
                graph=graph,
                language=language,
                _edge_index=str(edge_index),
                _parent=n_id,
                _parent_fields={
                    hash_node(child): fld
                    for fld in LANGUAGE_FIELDS[language].get(node.type, ())
                    for child in [node.child_by_field_name(fld)]
                    if child
                },
            )

    return graph


def parse_ast_graph(content: bytes, language: str) -> Graph | None:
    parser_language = PARSER_LANGUAGES[language]
    if not parser_language:
        LOGGER.warning("Unable to parse content for language %s", language)
        return None
    parser = Parser(parser_language)
    raw_tree: Tree = parser.parse(content)
    node: Node = raw_tree.root_node

    counter = map(str, count(1))
    try:
        return _build_ast_graph(
            content=content,
            node=node,
            counter=counter,
            graph=Graph(),
            language=language,
        )
    except ParsingError:
        return None


def adj_lazy(
    graph: Graph,
    n_id: NId,
    depth: int = 1,
) -> Iterator[NId]:
    """Return adjacent nodes to `n_id`, following just edges with given attrs.

    - Parameter `depth` may be -1 to indicate infinite depth.
    - Search is done breadth first.
    - Nodes are returned ordered ascending by index on each level.

    This function must be used instead of graph.adj, because graph.adj
    becomes unstable (unordered) after mutating the graph.
    """
    if depth == 0:
        return

    childs: list[str] = sorted(graph.adj[n_id], key=int)

    # Append direct childs
    yield from childs

    # Recurse into childs
    if depth < 0 or depth > 1:
        for c_id in childs:
            yield from adj_lazy(
                graph,
                c_id,
                depth=depth - 1,
            )


def adj(
    graph: Graph,
    n_id: NId,
    depth: int = 1,
) -> tuple[NId, ...]:
    return tuple(adj_lazy(graph, n_id, depth=depth))


def iter_childs(graph: Graph, n_id: str) -> Iterator[str]:
    for c_id in adj(graph, n_id):
        yield from iter_childs(graph, c_id)
    yield n_id


def lazy_text_childs(graph: Graph, n_id: str) -> Iterator[str]:
    for c_id in iter_childs(graph, n_id):
        if "label_text" in graph.nodes[c_id]:
            yield c_id


def lazy_childs_text(graph: Graph, n_id: str) -> Iterator[str]:
    for c_id in lazy_text_childs(graph, n_id):
        yield graph.nodes[c_id]["label_text"]


def node_to_str(graph: Graph, n_id: str, sep: str = "") -> str:
    return sep.join(lazy_childs_text(graph, n_id))



from packageurl import PackageURL

from labels.parsers.cataloger.java.model import JavaArchive, JavaPomProject, JavaPomProperties


def looks_like_group_id(group_id: str) -> bool:
    return "." in group_id


def remove_osgi_directives(group_id: str) -> str:
    return group_id.split(";")[0]


def clean_group_id(group_id: str) -> str:
    return remove_osgi_directives(group_id).strip()


def group_id_from_pom_properties(properties: JavaPomProperties | None) -> str:
    if not properties:
        return ""
    if properties.group_id:
        return clean_group_id(properties.group_id)
    if properties.artifact_id and looks_like_group_id(properties.artifact_id):
        return clean_group_id(properties.artifact_id)
    return ""


def group_id_pom_project(project: JavaPomProject | None) -> str:
    if not project:
        return ""
    if project.group_id:
        return clean_group_id(project.group_id)
    if project.artifact_id and looks_like_group_id(project.artifact_id):
        return clean_group_id(project.artifact_id)
    if project.parent:
        if project.parent.group_id:
            return clean_group_id(project.parent.group_id)
        if looks_like_group_id(project.parent.artifact_id):
            return clean_group_id(project.parent.artifact_id)
    return ""


def group_id_from_java_metadata(_pkg_name: str, metadata: JavaArchive | None) -> str | None:
    if metadata:
        if group_id := group_id_from_pom_properties(metadata.pom_properties):
            return group_id
        if group_id := group_id_pom_project(metadata.pom_project):
            return group_id
    return None


def package_url(name: str, version: str, metadata: JavaArchive | None = None) -> str:
    group_id = name
    if (g_id := group_id_from_java_metadata(name, metadata)) and g_id:
        group_id = g_id
    return PackageURL(
        type="maven",
        namespace=group_id,
        name=name,
        version=version,
        qualifiers=None,
        subpath="",
    ).to_string()



import glob
import os
import tarfile
from collections.abc import Generator
from pathlib import Path
from typing import TextIO, cast

from labels.model.file import Location, Metadata, Type
from labels.model.resolver import ContentResolver, FileReader, PathResolver, Resolver
from labels.model.sources import ImageContext, ImageMetadata
from labels.utils.file import new_location_from_image


def is_in_dependency_dir(path: str, dependency_dirs: set[str]) -> bool:
    for dep in dependency_dirs:
        dep_norm = dep if dep.endswith(os.path.sep) else dep + os.path.sep
        if path.startswith(dep_norm):
            return "node_modules" in path
    return False


def _normalize_rel_root(root: str, base_path: str) -> str:
    rel_root = os.path.relpath(root, base_path)
    return os.path.sep if rel_root == "." else os.path.join(os.path.sep, rel_root)


def get_type_from_tar_member(member: tarfile.TarInfo) -> Type:
    type_map = {
        member.issym: Type.TYPE_SYM_LINK,
        member.islnk: Type.TYPE_HARD_LINK,
        member.isreg: Type.TYPE_REGULAR,
        member.isdir: Type.TYPE_DIRECTORY,
        member.isfifo: Type.TYPE_FIFO,
        member.ischr: Type.TYPE_CHARACTER_DEVICE,
        member.isblk: Type.TYPE_BLOCK_DEVICE,
    }

    for check_func, type_value in type_map.items():
        if check_func():
            return type_value

    return Type.TYPE_IRREGULAR


class ContainerImagePathResolver(PathResolver):
    def __init__(self, *, img: ImageMetadata, context: ImageContext) -> None:
        self.img = img
        self.context = context

    def _has_path(self, path: str) -> bool:
        layer_ids = [x["digest"] for x in self.context.manifest["layers"]]

        success = []
        path = path.lstrip(os.path.sep)
        for layer_id in layer_ids:
            p_file_path = os.path.join(self.context.full_extraction_dir, layer_id, path)
            if Path(p_file_path).exists():
                success.append(True)
                break

        return any(success)

    def has_path(self, path: str) -> bool:
        return self._has_path(path)

    def _search_path(self, path: str) -> list[Location]:
        locations: list[Location] = []
        layer_ids = [x["digest"] for x in self.context.manifest["layers"]]

        for layer_id in layer_ids:
            p_file_path = os.path.join(
                self.context.full_extraction_dir,
                layer_id,
                path.lstrip(os.path.sep),
            )
            if Path(p_file_path).exists():
                locations.append(new_location_from_image(path, layer_id, p_file_path))
        return locations

    def files_by_path(self, *paths: str) -> list[Location]:
        locations: list[Location] = []

        for path in paths:
            if find_path := self._search_path(path.lstrip("/")):
                locations.extend(find_path)
            else:
                locations.extend(self._search_path(path))
        return locations

    def files_by_glob(self, *patterns: str) -> list[Location]:
        locations: list[Location] = []
        for layer_info in self.context.manifest["layers"]:
            layer_digest = layer_info["digest"]
            layer_extract_dir = os.path.join(self.context.full_extraction_dir, layer_digest)

            for pattern in patterns:
                full_pattern = os.path.join(layer_extract_dir, pattern)
                locations.extend(
                    new_location_from_image(
                        os.path.relpath(file_path, layer_extract_dir),
                        layer_digest,
                        file_path,
                    )
                    for file_path in glob.glob(full_pattern, recursive=True)  # noqa: PTH207
                )
        return locations

    def relative_file_path(self, _: Location, _path: str) -> Location | None:  # Review
        files = self.files_by_path(_path)
        if not files:
            return None
        return files[0]

    def _walk_layer(
        self,
        current_layer_path: str,
        global_dependency_dirs: set[str],
    ) -> Generator[str, None, None]:
        dependency_files = {"package.json", "package-lock.json", "yarn.lock"}
        for root, dirs, files in os.walk(current_layer_path):
            rel_root = _normalize_rel_root(root, current_layer_path)
            if is_in_dependency_dir(rel_root, global_dependency_dirs):
                continue
            if any(dep in files for dep in dependency_files):
                global_dependency_dirs.add(rel_root)
                if "node_modules" in dirs:
                    dirs.remove("node_modules")

            for file in files:
                yield os.path.join(rel_root, file)

    def walk_file(self) -> Generator[str, None, None]:
        layer_ids = [x["digest"] for x in self.context.manifest["layers"]]
        global_dependency_dirs: set[str] = set()

        for layer_id in layer_ids:
            current_layer_path = os.path.join(self.context.full_extraction_dir, layer_id)
            yield from self._walk_layer(current_layer_path, global_dependency_dirs)


class ContainerImageContentResolver(ContentResolver):
    def __init__(self, *, img: ImageMetadata, context: ImageContext) -> None:
        self.img = img
        self.context = context

    def file_contents_by_location(
        self,
        location: Location,
        *,
        function_reader: FileReader | None = None,
        mode: str | None = None,
    ) -> TextIO | None:
        if location.coordinates and Path(location.coordinates.real_path).exists():
            return cast(
                TextIO,
                (function_reader or open)(
                    location.coordinates.real_path,
                    encoding="utf-8",
                    mode=mode or "r",
                ),
            )
        return None


class ContainerImage(Resolver):
    img: ImageMetadata
    context: ImageContext
    _path_resolver: ContainerImagePathResolver
    _content_resolver: ContainerImageContentResolver

    def __init__(self, *, img: ImageMetadata, context: ImageContext) -> None:
        self.img = img
        self.context = context
        self._path_resolver = ContainerImagePathResolver(img=self.img, context=self.context)
        self._content_resolver = ContainerImageContentResolver(img=self.img, context=self.context)

    def has_path(self, path: str) -> bool:
        return self._path_resolver.has_path(path)

    def files_by_path(self, *paths: str) -> list[Location]:
        return self._path_resolver.files_by_path(*paths)

    def files_by_glob(self, *patterns: str) -> list[Location]:
        return self._path_resolver.files_by_glob(*patterns)

    def file_contents_by_location(
        self,
        location: Location,
        *,
        function_reader: FileReader | None = None,
        mode: str | None = None,
    ) -> TextIO | None:
        return self._content_resolver.file_contents_by_location(
            location,
            function_reader=function_reader,
            mode=mode,
        )

    def file_metadata_by_location(self, _location: Location) -> Metadata | None:
        raise NotImplementedError

    def relative_file_path(self, loc: Location, path: str) -> Location | None:
        return self._path_resolver.relative_file_path(loc, path)

    def walk_file(self) -> Generator[str, None, None]:
        yield from self._path_resolver.walk_file()



import abc
from collections.abc import Generator
from typing import Protocol, runtime_checkable


@runtime_checkable
class RpmDBInterfaceProtocol(Protocol):  # pragma: no cover
    def read(self) -> Generator[bytes, None, None]: ...


class RpmDBInterface(metaclass=abc.ABCMeta):  # pragma: no cover
    @classmethod
    def __subclasshook__(cls: type["RpmDBInterface"], subclass: object) -> bool:
        return bool(isinstance(subclass, RpmDBInterfaceProtocol))

    @abc.abstractmethod
    def read(
        self,
    ) -> Generator[bytes, None, None]:
        """Read entry bytes."""
        raise NotImplementedError



from typing import Any

from pydantic import BaseModel, ConfigDict


class AwsCredentials(BaseModel):
    access_key_id: str
    secret_access_key: str
    session_token: str | None


class AwsRole(BaseModel):
    external_id: str
    role: str


class LayerData(BaseModel):
    mimetype: str
    digest: str
    size: int
    annotations: dict[str, str] | None
    model_config = ConfigDict(frozen=True)


class ImageMetadata(BaseModel):
    name: str
    digest: str
    repotags: list[str]
    created: str
    dockerversion: str
    labels: dict[str, str] | None
    architecture: str
    os: str
    layers: list[str]
    layersdata: list[LayerData]
    env: list[str]
    image_ref: str = ""
    model_config = ConfigDict(frozen=True)


class LayerInfo(BaseModel):
    id_: str
    tar_full_path: str
    metadata: dict[str, Any]


class ImageContext(BaseModel):
    id: str
    name: str
    publisher: str | None
    arch: str
    size: str
    full_extraction_dir: str
    layers_dir: str
    manifest: dict[str, Any]
    image_ref: str



import logging
from copy import deepcopy
from typing import TYPE_CHECKING, cast

from packageurl import PackageURL
from pydantic import BaseModel, ValidationError
from urllib3.util import parse_url

from labels.model.relationship import Relationship
from labels.utils.strings import format_exception

if TYPE_CHECKING:  # pragma: no cover
    from labels.model.file import Location

from labels.model.file import DependencyType, LocationReadCloser
from labels.model.indexables import IndexedDict, Position
from labels.model.package import Language, Package, PackageType
from labels.model.resolver import Resolver
from labels.parsers.cataloger.generic.parser import Environment
from labels.parsers.collection.yaml import parse_yaml_with_tree_sitter


class DartPubspecLickEntry(BaseModel):
    name: str
    version: str
    hosted_url: str
    vcs_url: str


LOGGER = logging.getLogger(__name__)


def get_hosted_url(entry: IndexedDict[str, str | dict[str, str]]) -> str:
    hosted = entry.get("hosted")
    description: dict[str, str] | None = cast(
        dict[str, str] | None,
        entry.get("description"),
    )

    if hosted == "hosted" and description and description["url"] != "https://pub.dartlang.org":
        if host := parse_url(description["url"]).host:
            return host

        return description["url"]

    return ""


def get_vcs_url(entry: IndexedDict[str, str | dict[str, str]]) -> str:
    source = entry.get("source")
    description: dict[str, str] | None = cast(
        dict[str, str] | None,
        entry.get("description"),
    )

    if description and source == "git":
        if description.get("path") == ".":
            return f'{description["url"]}@{description["resolved-ref"]}'
        return description["url"] + f'@{description["resolved-ref"]}' + f'#{description["path"]}'
    return ""


def package_url(entry: DartPubspecLickEntry) -> str:
    qualifiers = {}
    if entry.hosted_url:
        qualifiers["hosted_url"] = entry.hosted_url
    elif entry.vcs_url:
        qualifiers["vcs_url"] = entry.vcs_url

    return PackageURL(  # type: ignore
        type="pub",
        namespace="",
        name=entry.name,
        version=entry.version,
        qualifiers=qualifiers,
        subpath="",
    ).to_string()


def parse_pubspec_lock(
    _: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    package_yaml = cast(
        IndexedDict[str, IndexedDict[str, IndexedDict[str, str | dict[str, str]]]],
        parse_yaml_with_tree_sitter(reader.read_closer.read()),
    )
    packages: list[Package] = []
    relationships: list[Relationship] = []
    yaml_packages: IndexedDict[str, IndexedDict[str, str | dict[str, str]]] = package_yaml[
        "packages"
    ]
    items = yaml_packages.items()

    for package_name, package_value in items:
        is_transitive = package_value.get("dependency") == "transitive"
        version = package_value.get("version")

        if not isinstance(version, str) or not package_name or not version:
            continue

        current_location: Location = deepcopy(reader.location)
        if current_location.coordinates:
            position: Position = yaml_packages.get_key_position(package_name)
            current_location.coordinates.line = position.start.line
            current_location.dependency_type = (
                DependencyType.TRANSITIVE if is_transitive else DependencyType.DIRECT
            )
        metadata = DartPubspecLickEntry(
            name=package_name,
            version=version,
            hosted_url=get_hosted_url(package_value),
            vcs_url=get_vcs_url(package_value),
        )
        try:
            packages.append(
                Package(
                    name=package_name,
                    version=version,
                    locations=[current_location],
                    language=Language.DART,
                    licenses=[],
                    type=PackageType.DartPubPkg,
                    p_url=package_url(metadata),
                    metadata=metadata,
                ),
            )
        except ValidationError as ex:
            LOGGER.warning(
                "Malformed package. Required fields are missing or data types are incorrect.",
                extra={
                    "extra": {
                        "exception": format_exception(str(ex)),
                        "location": current_location.path(),
                    },
                },
            )
            continue

    return packages, relationships



import re
from copy import deepcopy

from labels.model.file import DependencyType, Location, LocationReadCloser, Scope
from labels.model.package import Language, Package, PackageType
from labels.model.relationship import Relationship
from labels.model.resolver import Resolver
from labels.parsers.cataloger.elixir.utils import package_url
from labels.parsers.cataloger.generic.parser import Environment

MIX_LOCK_DEP: re.Pattern[str] = re.compile(r"^\"(?P<dep>[\w]*)\":\{(?P<info>.+)\},")


def build_location(
    reader: LocationReadCloser,
    line_number: int,
) -> Location:
    location = deepcopy(reader.location)
    location.scope = Scope.PROD
    if location.coordinates:
        location.coordinates.line = line_number
        location.dependency_type = DependencyType.DIRECT
    return location


def parse_mix_lock(
    _: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    packages: list[Package] = []
    relationships: list[Relationship] = []

    for line_number, line in enumerate(
        reader.read_closer.read().splitlines(),
        1,
    ):
        if matched := MIX_LOCK_DEP.match(line.replace(" ", "")):
            pkg_name = matched.group("dep")
            pkg_version = matched.group("info").split(",")[2].strip('"')
            location = build_location(reader, line_number)

            packages.append(
                Package(
                    name=pkg_name,
                    version=pkg_version,
                    type=PackageType.HexPkg,
                    locations=[location],
                    p_url=package_url(pkg_name, pkg_version),
                    metadata=None,
                    language=Language.ELIXIR,
                    licenses=[],
                    is_dev=False,
                ),
            )

    return packages, relationships



import atexit
import logging
import os
import sqlite3
from pathlib import Path
from typing import Literal

import boto3
import zstd
from botocore import UNSIGNED
from botocore.config import Config
from platformdirs import user_data_dir

LOGGER = logging.getLogger(__name__)


class BaseDatabase:
    BUCKET_NAME = "skims.sca"
    S3_SERVICE_NAME: Literal["s3"] = "s3"
    CONFIG_DIRECTORY = user_data_dir(
        appname="fluid-labels",
        appauthor="fluidattacks",
        ensure_exists=True,
    )

    def __init__(self, db_name: str) -> None:
        self.db_name = db_name
        self.bucket_file_key = f"{db_name}.zst"
        self.db_local_path = os.path.join(self.CONFIG_DIRECTORY, db_name)
        self.db_local_compressed_path = f"{self.db_local_path}.zst"
        self.connection: sqlite3.Connection | None = None
        self.s3_client = boto3.client(
            service_name=self.S3_SERVICE_NAME,
            config=Config(
                region_name="us-east-1",
                signature_version=UNSIGNED,
            ),
        )

    def _get_database_file(self) -> None:
        LOGGER.info("⬇️ Downloading advisories database")
        self.s3_client.download_file(
            Bucket=self.BUCKET_NAME,
            Key=self.bucket_file_key,
            Filename=self.db_local_compressed_path,
        )
        LOGGER.info("🗜️ Decompressing advisories database")

        try:
            with Path(self.db_local_compressed_path).open("rb") as compressed_file:
                compressed_data = compressed_file.read()
            uncompressed_data = zstd.decompress(compressed_data)
            with Path(self.db_local_path).open("wb") as output_file:
                output_file.write(uncompressed_data)
        except Exception:
            LOGGER.exception("❌ Unable to decompress database %s", self.db_name)

    def _initialize_db(self) -> bool:
        local_database_exists = Path(self.db_local_path).is_file()

        try:
            if self.is_up_to_date(local_database_exists=local_database_exists):
                LOGGER.info("✅ Advisories database is up to date")
                return True
            self._get_database_file()
            Path(self.db_local_compressed_path).unlink()
        except Exception:
            if local_database_exists:
                LOGGER.warning(
                    "⚠️ Advisories may be outdated, unable to update database",
                )
                return True

            LOGGER.exception(
                "❌ Advisories won't be included, unable to download database",
            )
            return False
        else:
            return True

    def is_up_to_date(self, *, local_database_exists: bool) -> bool:
        db_metadata = self.s3_client.head_object(
            Bucket=self.BUCKET_NAME,
            Key=self.bucket_file_key,
        )
        return (
            local_database_exists
            and Path(self.db_local_path).stat().st_mtime >= db_metadata["LastModified"].timestamp()
        )

    def initialize(self) -> None:
        if self.connection is None and self._initialize_db():
            self.connection = sqlite3.connect(
                self.db_local_path,
                check_same_thread=False,
            )
            atexit.register(self.connection.close)

    def get_connection(self) -> sqlite3.Connection:
        if self.connection is not None:
            return self.connection
        self.connection = sqlite3.connect(
            self.db_local_path,
            check_same_thread=False,
        )
        atexit.register(self.connection.close)
        return self.connection



import logging
from copy import deepcopy

from labels.model.file import LocationReadCloser
from labels.model.package import Package
from labels.model.relationship import Relationship
from labels.model.resolver import Resolver
from labels.parsers.cataloger.generic.parser import Environment
from labels.parsers.cataloger.graph_parser import Graph, NId, adj, node_to_str, parse_ast_graph
from labels.parsers.cataloger.swift.package import new_swift_package_manager_package

LOGGER = logging.getLogger(__name__)


def extract_package_name(url: str) -> str:
    return url.split("/")[-1].replace('"', "").replace("'", "").removesuffix(".git")


def extract_package_details(
    graph: Graph,
    package_node: NId,
) -> tuple[str | None, str | None, str | None]:
    package_name = None
    package_version = None
    source_url = None

    for argument_node in adj(graph, package_node, 3):
        if graph.nodes[argument_node].get("label_type") != "value_argument" or not (
            name_id := graph.nodes[argument_node].get("label_field_name")
        ):
            continue

        argument_name = node_to_str(graph, name_id)
        if argument_name == "url":
            source_url = node_to_str(graph, graph.nodes[argument_node]["label_field_value"])
            package_name = extract_package_name(source_url)
        elif argument_name in {"from", "version"}:
            package_version = (
                node_to_str(graph, graph.nodes[argument_node]["label_field_value"])
                .replace('"', "")
                .replace("'", "")
            )

    return package_name, package_version, source_url


def get_dependencies(
    dep_node: NId,
    graph: Graph,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    packages: list[Package] = []
    relationships: list[Relationship] = []

    for package_node in adj(graph, dep_node, depth=2):
        if is_valid_package_node(graph, package_node):
            package_name, package_version, source_url = extract_package_details(graph, package_node)
            if package_name and package_version:
                pkg = create_package(
                    graph,
                    package_node,
                    package_name,
                    package_version,
                    source_url,
                    reader,
                )
                if pkg:
                    packages.append(pkg)

    return packages, relationships


def is_valid_package_node(graph: Graph, package_node: NId) -> bool:
    return bool(
        graph.nodes[package_node]["label_type"] == "call_expression"
        and (child_value := adj(graph, package_node)[0])
        and node_to_str(graph, child_value) != ".package(",
    )


def create_package(  # noqa: PLR0913
    graph: Graph,
    package_node: NId,
    package_name: str,
    package_version: str,
    source_url: str | None,
    reader: LocationReadCloser,
) -> Package | None:
    new_location = deepcopy(reader.location)
    if new_location.coordinates:
        new_location.coordinates.line = graph.nodes[package_node]["label_l"]

    return new_swift_package_manager_package(
        name=package_name,
        version=package_version,
        source_url=source_url,
        location=new_location,
        revision=None,
    )


def parse_package_swift(
    _: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    content = reader.read_closer.read().encode("utf-8")
    dependencies = None
    if not content or not (graph := parse_ast_graph(content, "swift")):
        return ([], [])
    for node in graph.nodes:
        if (
            graph.nodes[node].get("label_type") == "call_expression"
            and (childs := adj(graph, node))
            and len(childs) == 2
            and graph.nodes[childs[0]].get("label_text") == "Package"
        ):
            for child in adj(graph, childs[1], depth=2):
                if (
                    graph.nodes[child].get("label_type") == "value_argument"
                    and (name_id := graph.nodes[child].get("label_field_name"))
                    and (name_value_id := adj(graph, name_id)[0])
                    and graph.nodes[name_value_id].get("label_text") == "dependencies"
                ):
                    dependencies = graph.nodes[child]["label_field_value"]
                    break
    if dependencies is None:
        return ([], [])
    return get_dependencies(dependencies, graph, reader)



import os
import uuid
from concurrent.futures import ThreadPoolExecutor
from datetime import UTC, datetime

from fluidattacks_tracks import Tracks
from fluidattacks_tracks.resources.event import Event

from labels.advisories.images import DATABASE as IMAGES_DATABASE
from labels.advisories.roots import DATABASE as ROOTS_DATABASE
from labels.config.bugsnag import initialize_bugsnag
from labels.config.logger import LOGGER, configure_logger, modify_logger_level
from labels.config.utils import guess_environment
from labels.core.source_dispatcher import resolve_sbom_source
from labels.enrichers.dispatcher import complete_package, complete_package_advisories_only
from labels.model.core import SbomConfig, SourceType
from labels.model.package import Package
from labels.model.relationship import Relationship
from labels.output.dispatcher import dispatch_sbom_output
from labels.parsers.operations.package_operation import package_operations_factory
from labels.resolvers.container_image import ContainerImage
from labels.resolvers.directory import Directory
from labels.utils.tracks import count_vulns_by_severity

client = Tracks()


def initialize_scan_environment(sbom_config: SbomConfig) -> None:
    configure_logger(log_to_remote=True)
    initialize_bugsnag()

    if sbom_config.debug:
        modify_logger_level()
    if sbom_config.source_type == SourceType.DIRECTORY:
        ROOTS_DATABASE.initialize()
    else:
        ROOTS_DATABASE.initialize()
        IMAGES_DATABASE.initialize()


def execute_labels_scan(sbom_config: SbomConfig) -> None:
    try:
        initialize_scan_environment(sbom_config)
        main_sbom_resolver = resolve_sbom_source(sbom_config)
        LOGGER.info(
            "📦 Generating SBOM from %s: %s",
            sbom_config.source_type.value,
            sbom_config.source,
        )

        packages, relationships = gather_packages_and_relationships(
            main_sbom_resolver,
            include_package_metadata=sbom_config.include_package_metadata,
        )

        LOGGER.info("📦 Preparing %s report", sbom_config.output_format.value)
        dispatch_sbom_output(
            packages=packages,
            relationships=relationships,
            config=sbom_config,
            resolver=main_sbom_resolver,
        )
        client.event.create(
            Event(
                action="CREATE",
                author="unknown",
                date=datetime.now(UTC),
                mechanism="TASK",
                metadata={
                    "status": "success",
                    "source": sbom_config.source,
                    "exclude": sbom_config.exclude,
                    "include": sbom_config.include,
                    "aws_role": sbom_config.aws_role,
                    "docker_user": sbom_config.docker_user,
                    "aws_external_id": sbom_config.aws_external_id,
                    "include_package_metadata": str(sbom_config.include_package_metadata),
                    "source_type": sbom_config.source_type.value,
                    "output_format": sbom_config.output_format.value,
                    "packages_count": str(len(packages)),
                    "relationships_count": str(len(relationships)),
                    "vulns_summary": count_vulns_by_severity(packages),
                },
                object="LabelsExecution",
                object_id=sbom_config.execution_id or str(uuid.uuid4()).replace("-", ""),
            ),
        )
    except Exception:
        if guess_environment() == "production":
            LOGGER.exception(
                "Error executing labels scan. Output SBOM was not generated.",
                extra={"execution_id": sbom_config.execution_id},
            )
            return
        raise


def gather_packages_and_relationships(
    resolver: Directory | ContainerImage,
    max_workers: int = 32,
    *,
    include_package_metadata: bool = True,
) -> tuple[list[Package], list[Relationship]]:
    packages, relationships = package_operations_factory(resolver)

    worker_count = min(
        max_workers,
        (os.cpu_count() or 1) * 5 if os.cpu_count() is not None else max_workers,
    )
    with ThreadPoolExecutor(max_workers=worker_count) as executor:
        LOGGER.info("📦 Gathering additional package information")
        if include_package_metadata:
            packages = list(filter(None, executor.map(complete_package, packages)))
        else:
            packages = list(executor.map(complete_package_advisories_only, packages))

    return packages, relationships
